import logging
import os

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.runtime import torch_device_fn
from flag_gems.utils import libentry, libtuner
from flag_gems.utils import triton_lang_extension as tle

from .utils import create_tma_device_descriptor, should_enable_sqmma

logger = logging.getLogger(
    f'flag_gems.runtime.backend._mthreads.ops.{__name__.split(".")[-1]}'
)


@triton.jit
def prev_multiple_of(a, b):
    # the largest x<a that x%b ==0
    return tl.cdiv(a, b) * b - b


@libentry()
@libtuner(
    configs=runtime.get_tuned_config("mm"),
    key=["M", "N", "K"],
    strategy=["align32", "align32", "align32"],
)
@triton.jit
def mm_kernel(
    A,
    B,
    C,
    M,
    N,
    K,
    stride_am,
    stride_ak,
    stride_bk,
    stride_bn,
    stride_cm,
    stride_cn,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
    BLOCK_K: tl.constexpr,
    GROUP_M: tl.constexpr,
):
    # matrix multiplication
    pid = tle.program_id(0)
    grid_m = tl.cdiv(M, BLOCK_M)
    grid_n = tl.cdiv(N, BLOCK_N)
    # re-order program ID for better L2 performance
    width = GROUP_M * grid_n
    group_id = pid // width
    group_size = min(grid_m - group_id * GROUP_M, GROUP_M)
    pid_m = group_id * GROUP_M + (pid % group_size)
    pid_n = (pid % width) // (group_size)
    # do matrix multiplication
    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    rn = pid_n * BLOCK_N + tl.arange(0, BLOCK_N)
    ram = tl.max_contiguous(tl.multiple_of(rm % M, BLOCK_M), BLOCK_M)
    rbn = tl.max_contiguous(tl.multiple_of(rn % N, BLOCK_N), BLOCK_N)
    prev_multiple = prev_multiple_of(K, BLOCK_K)

    acc = tl.zeros((BLOCK_M, BLOCK_N), dtype=tl.float32)
    for start_k in range(0, prev_multiple, BLOCK_K):
        rk = start_k + tl.arange(0, BLOCK_K)
        a = tl.load(A + (ram[:, None] * stride_am + rk[None, :] * stride_ak))
        b = tl.load(B + (rk[:, None] * stride_bk + rbn[None, :] * stride_bn))
        if a.dtype != b.dtype:
            a = a.to(C.dtype.element_ty)
            b = b.to(C.dtype.element_ty)
        acc += tl.dot(a, b, out_dtype=tl.float32, allow_tf32=False)

    # loop peeling
    rk = prev_multiple + tl.arange(0, BLOCK_K)
    mask_k = rk < K
    a = tl.load(
        A + (ram[:, None] * stride_am + rk[None, :] * stride_ak), mask=mask_k[None, :]
    )
    b = tl.load(
        B + (rk[:, None] * stride_bk + rbn[None, :] * stride_bn), mask=mask_k[:, None]
    )
    if a.dtype != b.dtype:
        a = a.to(C.dtype.element_ty)
        b = b.to(C.dtype.element_ty)
    acc += tl.dot(a, b, out_dtype=tl.float32, allow_tf32=False)

    acc = acc.to(C.dtype.element_ty)
    # rematerialize rm and rn to save registers
    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    rn = pid_n * BLOCK_N + tl.arange(0, BLOCK_N)
    C = C + (rm[:, None] * stride_cm + rn[None, :] * stride_cn)
    mask = (rm < M)[:, None] & (rn < N)[None, :]
    # handles write-back with reduction-splitting
    tl.store(C, acc, mask=mask)


_ordered_datatypes = [torch.float16, torch.bfloat16, torch.float32]


def get_higher_dtype(a, b):
    if a is b:
        return a

    assert a in _ordered_datatypes
    assert b in _ordered_datatypes

    for d in _ordered_datatypes:
        if a is d:
            return b
        if b is d:
            return a


def mm_fma(a, b):
    logger.debug("GEMS_MTHREADS MM(FMA)")
    device = a.device
    # handle non-contiguous inputs if necessary
    if a.stride(0) > 1 and a.stride(1) > 1:
        a = a.contiguous()
    if b.stride(0) > 1 and b.stride(1) > 1:
        b = b.contiguous()
    # checks constraints
    assert a.shape[1] == b.shape[0], "incompatible dimensions"
    M, K = a.shape
    _, N = b.shape
    # allocates output
    c_dtype = get_higher_dtype(a.dtype, b.dtype)
    c = torch.empty((M, N), device=device, dtype=c_dtype)
    # launch kernel
    grid = lambda META: (
        triton.cdiv(M, META["BLOCK_M"]) * triton.cdiv(N, META["BLOCK_N"]),
    )
    with torch_device_fn.device(a.device):
        mm_kernel[grid](
            a,
            b,
            c,
            M,
            N,
            K,
            a.stride(0),
            a.stride(1),
            b.stride(0),
            b.stride(1),
            c.stride(0),
            c.stride(1),
            GROUP_M=8,
        )
    return c


def mm_out(a, b, *, out):
    logger.debug("GEMS_MTHREADS MM_OUT")
    # handle non-contiguous inputs if necessary
    if a.stride(0) > 1 and a.stride(1) > 1:
        a = a.contiguous()
    if b.stride(0) > 1 and b.stride(1) > 1:
        b = b.contiguous()
    # checks constraints
    assert a.shape[1] == b.shape[0], "incompatible dimensions"
    M, K = a.shape
    _, N = b.shape
    # allocates output
    c = out
    # launch kernel
    grid = lambda META: (
        triton.cdiv(M, META["BLOCK_M"]) * triton.cdiv(N, META["BLOCK_N"]),
    )
    with torch_device_fn.device(a.device):
        mm_kernel[grid](
            a,
            b,
            c,
            M,
            N,
            K,
            a.stride(0),
            a.stride(1),
            b.stride(0),
            b.stride(1),
            c.stride(0),
            c.stride(1),
            GROUP_M=8,
        )
    return c


@triton.jit
def mm_sqmma_kernel(
    a_desc_ptr,
    b_desc_ptr,
    c_desc_ptr,
    M,
    N,
    K,
    GROUP_M: tl.constexpr,
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
    BLOCK_SIZE_K: tl.constexpr,
    ab_dtype: tl.constexpr,
    c_dtype: tl.constexpr,
    is_transpose_a: tl.constexpr = False,
    is_transpose_b: tl.constexpr = False,
):
    pid = tle.program_id(0)
    grid_m = tl.cdiv(M, BLOCK_SIZE_M)
    grid_n = tl.cdiv(N, BLOCK_SIZE_N)
    width = GROUP_M * grid_n
    group_id = pid // width
    group_size = min(grid_m - group_id * GROUP_M, GROUP_M)
    pid_m = group_id * GROUP_M + (pid % group_size)
    pid_n = (pid % width) // (group_size)
    offs_am = pid_m * BLOCK_SIZE_M
    offs_bn = pid_n * BLOCK_SIZE_N
    offs_k = 0
    offs_am = offs_am.to(tl.int32)
    offs_bn = offs_bn.to(tl.int32)
    offs_k = offs_k.to(tl.int32)
    accumulator = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)
    tme_load_ab_dtype = ab_dtype
    c_store_dtype = c_dtype
    for k in range(0, tl.cdiv(K, BLOCK_SIZE_K)):
        a = tl._experimental_descriptor_load(
            a_desc_ptr,
            [offs_am, offs_k],
            [BLOCK_SIZE_M, BLOCK_SIZE_K],
            tme_load_ab_dtype,
            is_transpose_a,
        )
        b = tl._experimental_descriptor_load(
            b_desc_ptr,
            [offs_k, offs_bn],
            [BLOCK_SIZE_K, BLOCK_SIZE_N],
            tme_load_ab_dtype,
            is_transpose_b,
        )
        accumulator += tl.dot(a, b, out_dtype=tl.float32, allow_tf32=False)
        offs_k += BLOCK_SIZE_K
    accumulator = accumulator.to(c_store_dtype)
    tl._experimental_descriptor_store(c_desc_ptr, accumulator, [offs_am, offs_bn])


def get_triton_type(elem_type):
    type_map = {
        torch.float16: tl.float16,
        torch.bfloat16: tl.bfloat16,
        torch.float8_e4m3fn: tl.float8e4nv,
    }
    return type_map.get(elem_type, None)


def mm_sqmma(A, B, M, N, K, GROUP_M, BLOCK_M, BLOCK_N, BLOCK_K, num_warps, num_stages):
    logger.debug("GEMS_MTHREADS MM(SQMMA)")
    device = "musa"
    # handle non-contiguous inputs if necessary
    is_transpose_a = False
    is_transpose_b = False
    if not A.is_contiguous():
        if A.stride(0) == 1 and A.stride(1) == A.shape[0]:
            is_transpose_a = True
        else:
            A = A.contiguous()
    if not B.is_contiguous():
        if B.stride(0) == 1 and B.stride(1) == B.shape[0]:
            is_transpose_b = True
        else:
            B = B.contiguous()
    a_type = A.dtype
    b_type = B.dtype
    assert a_type == b_type, "Mat A and Mat B should have the same dtype"
    c_dtype = get_higher_dtype(a_type, b_type)
    C = torch.empty((M, N), dtype=c_dtype, device=device)
    desc_a = create_tma_device_descriptor(A, BLOCK_M, BLOCK_K, device)
    desc_b = create_tma_device_descriptor(B, BLOCK_K, BLOCK_N, device)
    desc_c = create_tma_device_descriptor(C, BLOCK_M, BLOCK_N, device)
    mm_sqmma_kernel[(triton.cdiv(M, BLOCK_M) * triton.cdiv(N, BLOCK_N), 1, 1)](
        desc_a,
        desc_b,
        desc_c,
        M,
        N,
        K,
        GROUP_M,
        BLOCK_M,
        BLOCK_N,
        BLOCK_K,
        get_triton_type(a_type),
        get_triton_type(c_dtype),
        num_warps=num_warps,
        num_stages=num_stages,
        is_transpose_a=is_transpose_a,
        is_transpose_b=is_transpose_b,
    )
    return C


def mm(a, b):
    a_dtype = a.dtype
    b_dtype = b.dtype
    M, K = a.shape
    _, N = b.shape
    use_sqmma = should_enable_sqmma(a_dtype, b_dtype, M, N, K)
    if use_sqmma:
        GROUP_M = 8
        BLOCK_M = 128
        BLOCK_N = BLOCK_M
        BLOCK_K = 64
        num_warps = 16 if BLOCK_M == 256 else 4
        num_stages = 1
        return mm_sqmma(
            a,
            b,
            M,
            N,
            K,
            GROUP_M,
            BLOCK_M,
            BLOCK_N,
            BLOCK_K,
            num_warps,
            num_stages,
        )
    else:
        enable_sqmma = os.environ.pop("MUSA_ENABLE_SQMMA", None)
        result = mm_fma(a, b)
        if enable_sqmma:
            os.environ["MUSA_ENABLE_SQMMA"] = enable_sqmma
        return result




# Copyright (c) 2023-2025, Songlin Yang, Yu Zhang


import torch
import triton
import triton.language as tl

from fla.ops.utils import prepare_chunk_indices
from fla.ops.utils.op import exp


@triton.jit()
def chunk_update_once(
    b_p,
    b_k,
    b_v,
    b_m,
    b_g_exp_q,
    b_h,
    b_lamb,
):
    b_o = tl.dot((tl.dot(b_p.to(b_k.dtype), tl.trans(b_k)) * b_m).to(b_v.dtype), b_v)
    b_o += tl.dot((b_p * b_g_exp_q).to(b_h.dtype), b_h)
    if b_lamb is not None:
        b_o += b_lamb[None, :] * b_p
    return b_o


@triton.heuristics({
    'IS_VARLEN': lambda args: args['cu_seqlens'] is not None,
})
@triton.jit(do_not_specialize=['T'])
def chunk_fwd_mesa_cg_dim64_kernel(
    q,
    q_final,
    k,
    h,
    o,
    v,
    h_kv,
    g,
    beta,
    lamb,
    cu_seqlens,
    chunk_indices,
    T,
    max_CG_iteration: tl.constexpr,
    H: tl.constexpr,
    K: tl.constexpr,
    BT: tl.constexpr,
    BK: tl.constexpr,
    IS_VARLEN: tl.constexpr,
):
    i_t, i_bh = tl.program_id(0), tl.program_id(1)
    i_b, i_h = i_bh // H, i_bh % H

    if IS_VARLEN:
        i_tg = i_t
        i_n, i_t = tl.load(chunk_indices + i_t * 2).to(tl.int32), tl.load(chunk_indices + i_t * 2 + 1).to(tl.int32)
        bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(cu_seqlens + i_n + 1).to(tl.int32)
        T = eos - bos
        NT = tl.cdiv(T, BT)
    else:
        NT = tl.cdiv(T, BT)
        i_tg = i_b * NT + i_t
        bos, eos = i_b * T, i_b * T + T

    o_t = i_t * BT + tl.arange(0, BT)
    m_t = o_t < T

    # offset calculation
    q += (bos * H + i_h) * K
    q_final += (bos * H + i_h) * K
    k += (bos * H + i_h) * K
    h += (i_tg * H + i_h).to(tl.int64) * K * K
    g += bos * H + i_h
    beta += bos * H + i_h
    lamb += i_h * K

    o += (bos * H + i_h) * K
    v += (bos * H + i_h) * K
    h_kv += (i_tg * H + i_h).to(tl.int64) * K * K

    p_q = tl.make_block_ptr(q, (T, K), (H*K, 1), (i_t * BT, 0), (BT, BK), (1, 0))
    p_k = tl.make_block_ptr(k, (T, K), (H*K, 1), (i_t * BT, 0), (BT, BK), (1, 0))
    p_h = tl.make_block_ptr(h, (K, K), (K, 1), (0, 0), (BK, BK), (1, 0))

    b_h = tl.load(p_h, boundary_check=(0, 1))
    b_k = tl.load(p_k, boundary_check=(0, 1))
    b_q = tl.load(p_q, boundary_check=(0, 1)).to(tl.float32)

    p_g = tl.make_block_ptr(g, (T,), (H,), (i_t * BT,), (BT,), (0,))
    b_g = tl.load(p_g, boundary_check=(0,)).to(tl.float32)
    p_beta = tl.make_block_ptr(beta, (T,), (H,), (i_t * BT,), (BT,), (0,))
    b_beta = tl.load(p_beta, boundary_check=(0,)).to(tl.float32)
    p_lamb = tl.make_block_ptr(lamb, (K,), (1,), (0,), (BK,), (0,))

    b_lamb = tl.load(p_lamb, boundary_check=(0,)).to(tl.float32)

    b_m = exp(b_g[:, None] - b_g[None, :]) * b_beta[None, :]
    b_m = tl.where((o_t[:, None] >= o_t[None, :]) & (m_t[:, None] & m_t[None, :]), b_m, 0)
    b_g_exp_q = tl.exp(b_g)[:, None]

    b_x = tl.zeros([BT, BK], dtype=tl.float32)
    b_p = tl.zeros([BT, BK], dtype=tl.float32)
    b_r = tl.zeros([BT, BK], dtype=tl.float32)

    b_x += b_q * 0.
    b_r += b_q
    b_p += b_r
    b_delta_old = tl.sum(b_r*b_r, axis=1)
    for i in range(max_CG_iteration):
        b_o = chunk_update_once(b_p, b_k, b_k, b_m, b_g_exp_q, b_h, b_lamb)
        alpha = b_delta_old / (tl.sum(b_p*b_o, axis=1) + 1e-5)
        b_x += alpha[:, None] * b_p
        b_r = b_r - alpha[:, None] * b_o
        b_delta_new = tl.sum(b_r*b_r, axis=1)
        b_p = b_r + (b_delta_new / (b_delta_old + 1e-5))[:, None] * b_p
        b_delta_old = b_delta_new

    p_q_final = tl.make_block_ptr(q_final, (T, K), (H*K, 1), (i_t * BT, 0), (BT, BK), (1, 0))
    tl.store(p_q_final, b_x.to(p_q_final.dtype.element_ty), boundary_check=(0, 1))

    p_h_kv = tl.make_block_ptr(h_kv, (K, K), (K, 1), (0, 0), (BK, BK), (1, 0))
    b_h_kv = tl.load(p_h_kv, boundary_check=(0, 1))
    p_v = tl.make_block_ptr(v, (T, K), (H*K, 1), (i_t * BT, 0), (BT, BK), (1, 0))
    b_v = tl.load(p_v, boundary_check=(0, 1))
    b_o = chunk_update_once(b_x, b_k, b_v, b_m, b_g_exp_q, b_h_kv, None)
    p_o = tl.make_block_ptr(o, (T, K), (H*K, 1), (i_t * BT, 0), (BT, BK), (1, 0))
    tl.store(p_o, b_o.to(p_o.dtype.element_ty), boundary_check=(0, 1))


def chunk_mesa_cg_fwd(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    h: torch.Tensor,
    h_kv: torch.Tensor,
    g_local_cumsum: torch.Tensor,
    beta: torch.Tensor,
    lamb: torch.Tensor,  # lambda
    cu_seqlens: torch.Tensor | None = None,
    chunk_size: int = 64,
    max_CG_iteration: int = 30,
    output_dtype: torch.dtype | None = None,
) -> torch.Tensor:
    B, T, H, K = q.shape
    assert K <= 128, "head dimension must be less than 128"
    assert chunk_size <= 64 or K <= 64, "either chunk size or head dimension must be no greater than 64"
    q_final = torch.empty_like(q, dtype=q.dtype if output_dtype is None else output_dtype)

    assert v is not None, "v must be provided if calculate_output is True"
    assert h_kv is not None, "h_kv must be provided if calculate_output is True"
    o = torch.empty_like(v)

    chunk_indices = prepare_chunk_indices(cu_seqlens, chunk_size) if cu_seqlens is not None else None
    NT = triton.cdiv(T, chunk_size) if cu_seqlens is None else len(chunk_indices)
    BK = max(triton.next_power_of_2(K), 16)
    grid = (NT, H*B)

    chunk_fwd_mesa_cg_dim64_kernel[grid](
        q=q,
        q_final=q_final,
        o=o,
        v=v,
        h_kv=h_kv,
        k=k,
        h=h,
        g=g_local_cumsum,
        beta=beta,
        lamb=lamb,
        cu_seqlens=cu_seqlens,
        chunk_indices=chunk_indices,
        max_CG_iteration=max_CG_iteration,
        T=T,
        H=H,
        K=K,
        BT=chunk_size,
        BK=BK,
        num_warps=4,
        num_stages=1,
    )
    return q_final, o



"""
Tensor Parallel Execution Strategy

Distributes weight files across multiple GPUs via shard_map.
Execution uses CompiledSequence multi-device path with automatic
device alignment at op boundaries.

Functionally equivalent to FGP but selected when component-level
distribution (PP) doesn't work and block-level distribution is needed
at tensor granularity. The solver distributes weight shards evenly
across TP GPUs; CompiledSequence handles cross-device transfers.

Use case: Component too large for single GPU but topology supports
multi-GPU execution (NVLink or PCIe P2P).
"""

from typing import Dict, Any, Optional, List, Set
import torch

from .base import ExecutionStrategy, StrategyContext


class TensorParallelStrategy(ExecutionStrategy):
    """
    Tensor Parallel execution across multiple GPUs.

    Weight files distributed via shard_map (from Prism solver).
    Execution through CompiledSequence multi-device path.
    Device alignment handled by _run_inner_multi_device fast path.
    """

    def __init__(self, context: StrategyContext, strategy_name: str = "tp"):
        super().__init__(context, strategy_name)

        # Parse TP vs regular component allocations
        self.tp_components: Dict[str, List[str]] = {}
        self.regular_components: Dict[str, str] = {}

        for comp_name, alloc_info in context.allocations.items():
            if isinstance(alloc_info, dict):
                device_str = alloc_info.get('device', '')
            else:
                device_str = alloc_info[0] if isinstance(alloc_info, tuple) else str(alloc_info)

            if device_str.startswith("tp:"):
                devices = device_str[3:].split(",")
                self.tp_components[comp_name] = devices
            else:
                self.regular_components[comp_name] = device_str

        self._loaded_components: Set[str] = set()

    def execute_component(
        self,
        component_name: str,
        phase: str,
        inputs: Optional[Dict[str, Any]] = None,
    ) -> Optional[Dict[str, Any]]:
        """Execute component — TP or regular."""
        executor = self.context.component_executors.get(component_name)
        if executor is None:
            raise RuntimeError(
                f"ZERO FALLBACK: No executor for component '{component_name}'"
            )

        if component_name in self.tp_components:
            primary_device = self.tp_components[component_name][0]
        elif component_name in self.regular_components:
            primary_device = self.regular_components[component_name]
        else:
            primary_device = str(executor.device)

        # Transfer inputs to primary device
        if inputs:
            inputs = self.transfer_dict(inputs, primary_device, async_transfer=False)

        # Weights already loaded by RuntimeExecutor._ensure_weights_loaded() with shard_map.
        # CompiledSequence multi-device path handles cross-device execution.
        return executor.run(inputs or {})

    def prepare_inputs(
        self,
        component_name: str,
        inputs: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Transfer inputs to component's primary device."""
        if component_name in self.tp_components:
            target = self.tp_components[component_name][0]
        elif component_name in self.regular_components:
            target = self.regular_components[component_name]
        else:
            return inputs
        return self.transfer_dict(inputs, target, async_transfer=False)

    def handle_outputs(
        self,
        component_name: str,
        outputs: Dict[str, Any],
        target_device: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Transfer outputs to target device if needed."""
        if target_device is None:
            return outputs
        return self.transfer_dict(outputs, target_device, async_transfer=False)

    def get_component_device(self, component_name: str) -> str:
        """Get primary device for component."""
        if component_name in self.tp_components:
            return self.tp_components[component_name][0]
        elif component_name in self.regular_components:
            return self.regular_components[component_name]
        raise RuntimeError(f"ZERO FALLBACK: No device for '{component_name}'")

    def cleanup(self) -> None:
        """Release resources."""
        self._loaded_components.clear()



import torch
import triton
import triton.language as tl

from fla.ops.utils import prepare_chunk_indices


@triton.heuristics({
    'IS_VARLEN': lambda args: args['offsets'] is not None,
    'USE_GATE': lambda args: args['g_cumsum'] is not None,
})
@triton.jit(do_not_specialize=['T'])
def parallel_path_bwd_intra_chunk_kernel(
    q, k, v, g_cumsum, w1, w2,
    L, D,
    dq, dq_new, dk, dv, dw1, dw2, do, dg_cumsum,
    offsets, indices,
    T, scale,
    G: tl.constexpr, HQ: tl.constexpr, H: tl.constexpr,
    K: tl.constexpr, V: tl.constexpr, BK: tl.constexpr,  BV: tl.constexpr,
    BT: tl.constexpr, S: tl.constexpr,
    IS_VARLEN: tl.constexpr, USE_GATE: tl.constexpr,
):
    i_t, i_bh = tl.program_id(0), tl.program_id(1)
    i_b, i_hq = i_bh // HQ, i_bh % HQ
    i_h = i_hq // G

    if IS_VARLEN:
        i_n, i_t = tl.load(indices + i_t * 2).to(tl.int32), tl.load(indices + i_t * 2 + 1).to(tl.int32)
        bos, eos = tl.load(offsets + i_n).to(tl.int32), tl.load(offsets + i_n + 1).to(tl.int32)
        T = eos - bos
    else:
        i_n = i_b
        bos, eos = i_n * T, i_n * T + T

    # offset calculations
    k += (bos * H + i_h) * K  # GQA when H!=HQ
    v += (bos * H + i_h) * V  # GQA when H!=HQ
    w1 += (bos * H + i_h) * K
    w2 += (bos * H + i_h) * K

    q += (bos * HQ + i_hq) * K
    dq += (bos * HQ + i_hq) * K
    dq_new += (bos * HQ + i_hq) * K
    dk += (bos * HQ + i_hq) * K
    dv += (bos * HQ + i_hq) * V
    do += (bos * HQ + i_hq) * V
    dw1 += (bos * HQ + i_hq) * K
    dw2 += (bos * HQ + i_hq) * K
    L += (bos * HQ + i_hq)
    D += (bos * HQ + i_hq)
    if USE_GATE:
        g_cumsum += (bos * HQ + i_hq)
        dg_cumsum += (bos * HQ + i_hq)

    # constants
    sm_scale = scale * 1.44269504

    p_do = tl.make_block_ptr(do, (T, V), (HQ*V, 1), (i_t * BT, 0), (BT, BV), (1, 0))
    # [BT, BV]
    b_do = tl.load(p_do, boundary_check=(0, 1))

    p_delta = tl.make_block_ptr(D, (T, ), (HQ, ), (i_t * BT, ), (BT, ), (0, ))
    b_delta = tl.load(p_delta, boundary_check=(0, ))
    p_l = tl.make_block_ptr(L, (T, ), (HQ, ), (i_t * BT, ), (BT, ), (0, ))
    b_l = tl.load(p_l, boundary_check=(0, ))

    b_dq = tl.zeros([BT, BK], dtype=tl.float32)
    p_dq = tl.make_block_ptr(dq, (T, K), (HQ*K, 1), (i_t * BT, 0), (BT, BK), (1, 0))
    b_dq += tl.load(p_dq, boundary_check=(0, 1))
    p_q = tl.make_block_ptr(q, (T, K), (HQ*K, 1), (i_t * BT, 0), (BT, BK), (1, 0))
    b_q = tl.load(p_q, boundary_check=(0, 1))

    if USE_GATE:
        p_gq_cumsum = tl.make_block_ptr(g_cumsum, (T, ), (HQ, ), (i_t * BT, ), (BT, ), (0, ))
        b_gq_cumsum = tl.load(p_gq_cumsum, boundary_check=(0, ))
        b_dgq = tl.zeros([BT ], dtype=tl.float32)
    else:
        b_dgq = None

    curr_start = (tl.floor(i_t * BT / S).to(tl.int32) * S).to(tl.int32)

    for offset in range(curr_start, i_t * BT, BT):
        mask = offset + tl.arange(0, BT) < T
        p_k = tl.make_block_ptr(k, (T, K), (H*K, 1), (offset, 0), (BT, BK), (1, 0))
        b_k = tl.load(p_k, boundary_check=(0, 1))
        b_q_tmp = tl.zeros([BT, BK], dtype=tl.float32)
        b_q_tmp += b_q
        for i_t_small in range(i_t * BT - BT, offset, -BT):
            p_w1 = tl.make_block_ptr(w1, (T, K), (H*K, 1), (i_t_small, 0), (BT, BK), (1, 0))
            b_w1 = tl.load(p_w1, boundary_check=(0, 1))
            p_w2 = tl.make_block_ptr(w2, (T, K), (H*K, 1), (i_t_small, 0), (BT, BK), (1, 0))
            b_w2 = tl.load(p_w2, boundary_check=(0, 1))
            b_A_tmp = tl.dot(b_q_tmp.to(b_w1.dtype), tl.trans(b_w1))
            b_q_tmp -= tl.dot(b_A_tmp.to(b_w1.dtype), b_w2)
        b_q2 = b_q_tmp.to(b_k.dtype)
        b_A = tl.dot(b_q2, tl.trans(b_k))
        if USE_GATE:
            p_gk_cumsum = tl.make_block_ptr(g_cumsum, (T, ), (HQ, ), (offset, ), (BT, ), (0, ))
            b_gk_cumsum = tl.load(p_gk_cumsum, boundary_check=(0, ))
            b_A = b_A + b_gq_cumsum[:, None] - b_gk_cumsum[None, :]
            b_A = tl.where((i_t * BT + tl.arange(0, BT) < T)[:, None], b_A, float("-inf"))  # avoid nan
        b_A_softmax = tl.math.exp2(b_A * sm_scale - b_l[:, None])
        b_dv = tl.dot(tl.trans(b_A_softmax.to(b_do.dtype)), b_do)
        tl.atomic_add(
            dv + ((offset + tl.arange(0, BT)) * HQ * V)[:, None] + tl.arange(0, BV)[None, :],
            b_dv.to(dv.dtype.element_ty),
            mask=mask[:, None],
            sem='relaxed',
        )
        p_v = tl.make_block_ptr(v, (T, V), (V*H, 1), (offset, 0), (BT, BV), (1, 0))
        b_v = tl.load(p_v, boundary_check=(0, 1))
        b_dp = tl.dot(b_do, tl.trans(b_v))
        b_dA = ((b_dp - b_delta[:, None]) * b_A_softmax * scale)
        if USE_GATE:
            b_dgk = -tl.sum(b_dA, axis=0)
            tl.atomic_add(dg_cumsum + (offset + tl.arange(0, BT)) * HQ, b_dgk, mask=mask, sem='relaxed')
            b_dgq += tl.sum(b_dA, axis=1)
        b_dA = b_dA.to(b_v.dtype)
        b_dk = tl.dot(tl.trans(b_dA), b_q2)
        tl.atomic_add(dk + (offset + tl.arange(0, BT))[:, None] * HQ*K + tl.arange(0,
                      BK)[None, :], b_dk, mask=mask[:, None], sem='relaxed')
        p_w1 = tl.make_block_ptr(w1, (T, K), (H*K, 1), (offset, 0), (BT, BK), (1, 0))
        b_w1 = tl.load(p_w1, boundary_check=(0, 1))
        p_w2 = tl.make_block_ptr(w2, (T, K), (H*K, 1), (offset, 0), (BT, BK), (1, 0))
        b_w2 = tl.load(p_w2, boundary_check=(0, 1))
        b_dA2 = tl.dot(b_dq.to(b_w2.dtype), tl.trans(b_w2)).to(b_v.dtype)
        b_A2 = tl.dot(b_q2.to(b_w1.dtype), tl.trans(b_w1)).to(b_v.dtype)
        b_dw2 = -tl.dot(tl.trans(b_A2), b_dq.to(b_v.dtype))
        tl.atomic_add(dw2 + (offset + tl.arange(0, BT))[:, None] * HQ*K + tl.arange(0,
                      BK)[None, :], b_dw2, mask=mask[:, None], sem='relaxed')
        b_dw1 = -tl.dot(tl.trans(b_dA2), b_q2.to(b_v.dtype))
        tl.atomic_add(dw1 + (offset + tl.arange(0, BT))[:, None] * HQ*K + tl.arange(0,
                      BK)[None, :], b_dw1, mask=mask[:, None], sem='relaxed')
        b_dq -= tl.dot(b_dA2, b_w1.to(b_v.dtype))
        b_dq += tl.dot(b_dA.to(b_k.dtype), b_k)

    p_dq_new = tl.make_block_ptr(dq_new, (T, K), (HQ*K, 1), (i_t * BT, 0), (BT, BK), (1, 0))
    tl.store(p_dq_new, b_dq.to(dq_new.dtype.element_ty), boundary_check=(0, 1))
    mask = i_t * BT + tl.arange(0, BT) < T
    if USE_GATE:
        tl.atomic_add(dg_cumsum + (i_t * BT + tl.arange(0, BT)) * HQ, b_dgq, mask=mask, sem='relaxed')


def parallel_path_bwd_intra_chunk_fn(
    q, k, v, g_cumsum, w1, w2,
    dq, dk, dv, dg_cumsum, dw1, dw2, do,
    scale, L, D,
    cu_seqlens,
    S, BT,
):
    assert dk.dtype == dv.dtype == dw1.dtype == dw2.dtype == torch.float32, 'atomic_add requires float32'
    B, T, HQ, K = q.shape
    assert dk.shape == dq.shape

    V = v.shape[-1]
    H = k.shape[-2]
    G = HQ // H
    indices = prepare_chunk_indices(cu_seqlens, BT) if cu_seqlens is not None else None
    NT = triton.cdiv(T, BT) if cu_seqlens is None else len(indices)
    dq_new = torch.empty_like(dq, dtype=q.dtype)
    parallel_path_bwd_intra_chunk_kernel[(NT, B*HQ)](
        q=q, k=k, v=v, g_cumsum=g_cumsum,
        w1=w1, w2=w2, L=L, D=D,
        dq=dq, dq_new=dq_new, dk=dk, dv=dv, dw1=dw1, dw2=dw2,
        do=do, dg_cumsum=dg_cumsum,
        offsets=cu_seqlens, indices=indices,
        T=T, S=S, BT=BT, scale=scale,
        G=G, HQ=HQ, H=H, K=K, V=V,
        BK=triton.next_power_of_2(K), BV=triton.next_power_of_2(V),
    )
    return dq_new, dk, dv, dw1, dw2, dg_cumsum



"""
RoPE Kernel - NVIDIA
Rotary Position Embedding

Source: Liger-Kernel (adapted)
Tier: common
"""
import torch
import triton
import triton.language as tl

from neurobrix.kernels.registry import register_kernel


# =============================================================================
# TRITON KERNELS
# =============================================================================

@triton.jit
def _rope_kernel(
    q_ptr,
    k_ptr,
    cos_ptr,
    sin_ptr,
    q_out_ptr,
    k_out_ptr,
    stride_q_row, stride_q_head, stride_q_dim,
    stride_k_row, stride_k_head, stride_k_dim,
    stride_cos_row, stride_cos_dim,
    stride_sin_row, stride_sin_dim,
    stride_q_out_row, stride_q_out_head, stride_q_out_dim,
    stride_k_out_row, stride_k_out_head, stride_k_out_dim,
    n_rows, n_heads_q, n_heads_k, head_dim,
    BLOCK_SIZE: tl.constexpr,
):
    """
    Apply RoPE to Q and K.
    Grid: (n_rows * max(n_heads_q, n_heads_k))
    """
    pid = tl.program_id(0)
    
    # We parallelize over (batch * seq_len) * heads
    # But n_heads_q and n_heads_k might differ (GQA/MQA)
    # For simplicity here, assuming separate kernels or handling max
    
    # Simplified version: 1 block per token per head
    # pid = row_idx * n_heads + head_idx
    
    # Let's assume standard RoPE where we apply same rotation to all heads
    # Each thread block handles one head vector of size head_dim
    
    # TODO: This is a simplified placeholder. RoPE is complex due to different shapes.
    # For Phase 3.2, we implement a naive version that assumes Q and K have same number of heads
    # or we just process Q and K separately.
    
    # Let's process just Q for now as a building block
    # pid corresponds to (row, head) tuple flattened
    
    # row_idx = pid // n_heads_q
    # head_idx = pid % n_heads_q
    
    # offsets = tl.arange(0, head_dim)
    pass 


# =============================================================================
# WRAPPER
# =============================================================================

def _rope_impl(q: torch.Tensor, k: torch.Tensor, cos: torch.Tensor, sin: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Apply Rotary Position Embedding.
    
    q, k: [batch, seq_len, num_heads, head_dim]
    cos, sin: [seq_len, head_dim] OR [batch, seq_len, head_dim]
    """
    # Fallback to PyTorch for complex broadcasting logic until Triton kernel is fully ported
    # Liger's implementation is quite complex with many strides
    
    # Standard RoPE implementation in PyTorch (efficient enough for now)
    # This acts as a reference implementation while we prepare the optimized Triton one
    
    # Reshape for broadcasting
    # q: [batch, seq_len, num_heads, head_dim]
    # cos: [seq_len, head_dim] -> [1, seq_len, 1, head_dim]
    
    if cos.ndim == 2:
        cos = cos.unsqueeze(0).unsqueeze(2)
        sin = sin.unsqueeze(0).unsqueeze(2)
    elif cos.ndim == 3:
        cos = cos.unsqueeze(2)
        sin = sin.unsqueeze(2)
        
    def rotate_half(x):
        x1, x2 = x[..., :x.shape[-1]//2], x[..., x.shape[-1]//2:]
        return torch.cat((-x2, x1), dim=-1)

    q_embed = (q * cos) + (rotate_half(q) * sin)
    k_embed = (k * cos) + (rotate_half(k) * sin)
    
    return q_embed, k_embed


# =============================================================================
# REGISTRATION
# =============================================================================

@register_kernel(family="llm", vendor="nvidia", tier="common", op_name="rope")
def rope_kernel(q: torch.Tensor, k: torch.Tensor, cos: torch.Tensor, sin: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
    """RoPE kernel compatible with all NVIDIA architectures."""
    return _rope_impl(q, k, cos, sin)



import logging
import math

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.runtime import torch_device_fn
from flag_gems.utils import libentry
from flag_gems.utils import triton_lang_extension as tle

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))
torch_dtype_to_tl_dtype_and_max_value = {
    torch.int16: (tl.int16, torch.iinfo(torch.int16).max),
    torch.int32: (tl.int32, torch.iinfo(torch.int32).max),
    torch.float16: (tl.float16, torch.finfo(torch.float16).max),
    torch.float32: (tl.float32, torch.finfo(torch.float32).max),
    torch.bfloat16: (tl.float32, torch.finfo(torch.float32).max),
}


@libentry()
@triton.jit
def argmin_kernel_1(
    inp,
    mid_value,
    mid_index,
    M,
    BLOCK_SIZE: tl.constexpr,
    dtype_max_value: tl.constexpr,
):
    pid = tle.program_id(0)
    offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    inp_ptrs = inp + offset
    mask = offset < M
    inp_val = tl.load(inp_ptrs, mask=mask, other=dtype_max_value)
    min_val, min_index = tl.min(inp_val, axis=0, return_indices=True)
    min_index = min_index + pid * BLOCK_SIZE
    mid_value_ptr = mid_value + pid
    min_index_ptr = mid_index + pid
    tl.store(mid_value_ptr, min_val)
    tl.store(min_index_ptr, min_index)


@libentry()
@triton.jit
def argmin_kernel_2(
    mid_value,
    mid_index,
    out,
    mid_size,
    BLOCK_MID: tl.constexpr,
    dtype_max_value: tl.constexpr,
):
    offset = tl.arange(0, BLOCK_MID)
    mid_ptrs = mid_value + offset
    mask = offset < mid_size
    mid_val = tl.load(mid_ptrs, mask=mask, other=dtype_max_value)
    index_val = tl.argmin(mid_val, axis=0)
    mid_index_ptrs = mid_index + index_val
    out_val = tl.load(mid_index_ptrs)
    tl.store(out, out_val)


def heur_block_n(args):
    return min(4096, triton.next_power_of_2(args["N"]))


@libentry()
@triton.heuristics(runtime.get_heuristic_config("argmin"))
@triton.jit
def argmin_kernel(
    inp,
    out_index,
    M,
    N,
    K,
    tl_dtype: tl.constexpr,
    dtype_max_value: tl.constexpr,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    # set offset
    pid_m = tle.program_id(0)
    pid_k = tle.program_id(1)
    m_offset = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)

    # min_values = tl.full([BLOCK_M], dtype=tl.float32, value=float("inf"))
    if tl_dtype is tl.int16:
        tl_dtype = tl.int32
    n_offset = tl.arange(0, BLOCK_N)
    offset = m_offset[:, None] * N * K + n_offset[None, :] * K + pid_k
    offset_index = m_offset * K + pid_k
    # set mask
    mask1 = m_offset < M
    mask = m_offset[:, None] < M and n_offset[None, :] < N
    inp_ptrs = inp + offset
    inp_vals = tl.load(inp_ptrs, mask=mask, other=-float("inf"))
    # inp_vals = tl.where(mask, inp_vals, -float("inf"))
    _, result_index = tl.min(inp_vals, axis=1, return_indices=True)

    out_index_ptrs = out_index + offset_index

    tl.store(out_index_ptrs, result_index, mask=mask1)


def argmin(inp, dim=None, keepdim=False, *, dtype=None):
    logger.debug("GEMS argmin")
    if dim is None:
        M = inp.numel()
        if dtype is None:
            dtype = inp.dtype
        block_size = triton.next_power_of_2(math.ceil(math.sqrt(M)))
        mid_size = triton.cdiv(M, block_size)
        block_mid = triton.next_power_of_2(mid_size)

        mid_value = torch.empty((mid_size,), dtype=dtype, device=inp.device)
        mid_index = torch.empty((mid_size,), dtype=torch.int64, device=inp.device)
        if keepdim:
            shape = list(inp.shape)
            for i in range(0, inp.dim()):
                shape[i] = 1
            out = torch.empty(shape, dtype=torch.int64, device=inp.device)
        else:
            out = torch.empty([], dtype=torch.int64, device=inp.device)

        tl_dtype, dtype_max_value = torch_dtype_to_tl_dtype_and_max_value[inp.dtype]
        with torch_device_fn.device(inp.device):
            argmin_kernel_1[(mid_size, 1, 1)](
                inp,
                mid_value,
                mid_index,
                M,
                block_size,
                dtype_max_value,
            )
            argmin_kernel_2[(1, 1, 1)](
                mid_value,
                mid_index,
                out,
                mid_size,
                block_mid,
                dtype_max_value,
            )
        return out
    else:
        assert dim >= -inp.ndim and dim < inp.ndim, "Invalid dim"
        shape = inp.shape
        dim = dim % inp.ndim
        N = shape[dim]
        M = math.prod(shape[:dim])
        K = inp.numel() // M // N

        inp = inp.contiguous()

        shape_list = list(shape)
        shape_list[dim] = 1
        out_index = torch.empty(shape_list, dtype=torch.int64, device=inp.device)
        if not keepdim:
            out_index = torch.squeeze(out_index, dim)

        tl_dtype, dtype_max_value = torch_dtype_to_tl_dtype_and_max_value[inp.dtype]

        grid = lambda meta: (
            triton.cdiv(M, meta["BLOCK_M"]),
            K,
        )
        with torch_device_fn.device(inp.device):
            argmin_kernel[grid](
                inp,
                out_index,
                M,
                N,
                K,
                tl_dtype,
                dtype_max_value,
            )

        return out_index



#  Copyright 2022 Lefebvre Sarrut
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from typing import Optional, Union

import torch
import triton
import triton.language as tl
from torch.autograd.function import FunctionCtx
from torch.cuda.amp import custom_fwd


def attention_split_1_reference(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    sm_scale: float,
    is_causal: bool,
    attention_mask: Union[torch.Tensor, None],
) -> torch.Tensor:
    BLOCK_N = 128
    size_n = k.size(2)
    n_divisions = triton.cdiv(size_n, BLOCK_N)

    p = torch.matmul(q.float(), k.transpose(2, 3).float()) * sm_scale

    if attention_mask is not None:
        p += attention_mask
    if is_causal:
        size_m = q.size(2)
        size_n = k.size(2)
        M = torch.tril(torch.ones((size_m, size_n), device="cuda"))
        p = torch.where(M == 0, float("-inf"), p)

    p = p.view(p.size(0), p.size(1), p.size(2), n_divisions, p.size(3) // n_divisions).permute(0, 1, 3, 2, 4)

    maximums, _ = torch.max(p, 4)
    numerators = torch.exp(p - maximums.view(*maximums.size(), 1))
    sums = torch.sum(numerators, 4)

    v = v.view(v.size(0), v.size(1), n_divisions, v.size(2) // n_divisions, v.size(3))
    blocks = torch.matmul(numerators.half(), v)

    return blocks, maximums, sums


@triton.jit
def _fwd_part_1(
    head_size,
    m_size,
    n_size,
    q_ptr,
    k_ptr,
    v_ptr,
    sm_scale,
    attention_mask_ptr,
    output_ptr,
    maximums_ptr,
    sums_ptr,
    q_batch_stride,
    q_head_stride,
    q_m_stride,
    q_k_stride,
    k_batch_stride,
    k_head_stride,
    k_n_stride,
    k_k_stride,  # axis named n,k instead of k,n because of the transpose of K matrix
    v_batch_stride,
    v_head_stride,
    v_k_stride,
    v_n_stride,
    sums_batch_stride,
    sums_head_stride,
    sums_step_stride,
    sums_m_stride,
    maximums_batch_stride,
    maximums_head_stride,
    maximums_step_stride,
    maximums_m_stride,
    output_batch_stride,
    output_head_stride,
    output_step_stride,
    output_m_stride,
    output_n_stride,
    attention_mask_batch_stride,
    attention_mask_head_stride,
    attention_mask_m_stride,
    attention_mask_k_stride,
    min_clamp_value,
    N_LOAD_MASK_NEEDED: tl.constexpr,
    M_LOAD_MASK_NEEDED: tl.constexpr,
    MASK_BATCH_SIZE: tl.constexpr,
    MASK_HEAD_SIZE: tl.constexpr,
    MASK_M_SIZE: tl.constexpr,
    MASK_K_SIZE: tl.constexpr,
    HAS_MASK: tl.constexpr,
    IS_CAUSAL: tl.constexpr,
    BLOCK_M_SIZE: tl.constexpr,
    BLOCK_DHEAD_SIZE: tl.constexpr,
    BLOCK_N_SIZE: tl.constexpr,
):
    # Index of the block on M axis (M axis is the rows of matrix K)
    block_n_idx = tl.program_id(0)
    block_m_idx = tl.program_id(1)
    head_idx = tl.program_id(2)

    # offsets
    m_range_offs = tl.arange(0, BLOCK_M_SIZE)  # first block on M dimension
    n_range_offs = tl.arange(0, BLOCK_N_SIZE)  # first block on N dimension
    d_range_offs = tl.arange(0, BLOCK_DHEAD_SIZE)  # full head

    m_offs = block_m_idx * BLOCK_M_SIZE + m_range_offs  # rows offsets on M axis

    current_batch_idx = head_idx // head_size
    current_head_idx = head_idx % head_size
    # memory offsets matrices on whole Q, K, V matrices
    # Offsets for the current block on matrix Q
    q_offs = (
        current_batch_idx * q_batch_stride
        + current_head_idx * q_head_stride
        + (m_offs[:, None] * q_m_stride + d_range_offs[None, :] * q_k_stride)
    )

    # Offsets for the first block on matrix K
    k_offs = (
        current_batch_idx * k_batch_stride
        + current_head_idx * k_head_stride
        + (n_range_offs[:, None] * k_n_stride + d_range_offs[None, :] * k_k_stride)
    )

    # Offsets for the first block on matrix V
    v_offs = (
        current_batch_idx * v_batch_stride
        + current_head_idx * v_head_stride
        + (n_range_offs[:, None] * v_k_stride + d_range_offs[None, :] * v_n_stride)
    )

    # pointers to blocks in Q, K, V
    q_ptrs = q_ptr + q_offs
    k_ptrs = k_ptr + k_offs
    v_ptrs = v_ptr + v_offs

    # load q, a block of full rows of matrix q
    # it will stay in SRAM throughout
    if M_LOAD_MASK_NEEDED:
        q = tl.load(q_ptrs, mask=m_offs[:, None] < m_size, eviction_policy="", other=0.0)
    else:
        q = tl.load(q_ptrs, eviction_policy="")

    if HAS_MASK:
        attention_mask_batch_idx = (current_batch_idx,)
        if MASK_BATCH_SIZE == 1:
            attention_mask_batch_idx = 0

        attention_mask_head_idx = current_head_idx
        if MASK_HEAD_SIZE == 1:
            attention_mask_head_idx = 0

        attention_mask_off = (
            attention_mask_batch_idx * attention_mask_batch_stride
            + attention_mask_head_idx * attention_mask_head_stride
        )

    block_n_start_idx = block_n_idx * BLOCK_N_SIZE
    block_n_offs = block_n_start_idx + n_range_offs

    # We load the current block in K in SRAM
    # We do the first multiplication between the block in Q and the current block in K
    # We finish with the scaling (sqrt(BLOCK_DHEAD) in Vaswani et al. but sm_scale here)
    if N_LOAD_MASK_NEEDED:
        k_ptr_mask = block_n_offs[:, None] < n_size
        k = tl.load(k_ptrs + block_n_start_idx * k_n_stride, mask=k_ptr_mask, eviction_policy="", other=0.0)
    else:
        k = tl.load(k_ptrs + block_n_start_idx * k_n_stride, eviction_policy="")

    qk = tl.zeros((BLOCK_M_SIZE, BLOCK_N_SIZE), dtype=tl.float32)

    # required to fix a Triton compiler bug, if not done, there is a precision issue
    if N_LOAD_MASK_NEEDED:
        qk = tl.where(n_range_offs[None, :] < n_size, qk, float("-inf"))
    qk += tl.dot(q, tl.trans(k))
    qk *= sm_scale
    if IS_CAUSAL:
        qk += tl.where(m_offs[:, None] >= block_n_offs[None, :], 0, float("-inf"))

    if HAS_MASK:
        # we assume mask has a vector shape
        attention_mask_offs = attention_mask_off + block_n_offs[None, :] * attention_mask_k_stride
        if MASK_M_SIZE != 1:  # mask has a square shape, we load (BLOCK_M, BLOCK_N) elements
            attention_mask_offs += m_offs[:, None] * attention_mask_m_stride

        if N_LOAD_MASK_NEEDED & MASK_M_SIZE == 1:  # mask has a vector shape need a load mask
            attention_mask_ptr_mask = block_n_offs[None, :] < n_size
        if MASK_M_SIZE != 1:  # mask has a matrix shape
            if M_LOAD_MASK_NEEDED & (not N_LOAD_MASK_NEEDED):  # load mask on M axis
                attention_mask_ptr_mask = m_offs[:, None] < m_size
            elif (not M_LOAD_MASK_NEEDED) & N_LOAD_MASK_NEEDED:  # load mask on N axis
                attention_mask_ptr_mask = block_n_offs[None, :] < n_size
            elif M_LOAD_MASK_NEEDED & N_LOAD_MASK_NEEDED:  # load mask on both axis
                attention_mask_ptr_mask = (block_n_offs[None, :] < n_size) & (m_offs[:, None] < m_size)

        if M_LOAD_MASK_NEEDED | N_LOAD_MASK_NEEDED:
            attention_mask = tl.load(
                attention_mask_ptr + attention_mask_offs,
                eviction_policy="",
                mask=attention_mask_ptr_mask,
                other=float("-inf"),
            )
        else:
            attention_mask = tl.load(
                attention_mask_ptr + attention_mask_offs,
                eviction_policy="",
            )
        # Avoids NaN
        attention_mask = tl.where(attention_mask == float("-inf"), min_clamp_value, attention_mask)
        qk += attention_mask

    l_j = tl.max(qk, 1)
    numerators = tl.exp(qk - l_j[:, None])
    d_j = tl.sum(numerators, 1)

    maximums_offs = (
        current_batch_idx * maximums_batch_stride
        + current_head_idx * maximums_head_stride
        + block_n_idx * maximums_step_stride
        + m_offs * maximums_m_stride
    )
    maximums_ptrs = maximums_ptr + maximums_offs
    tl.store(maximums_ptrs, l_j, mask=m_offs < m_size)

    sums_offs = (
        current_batch_idx * sums_batch_stride
        + current_head_idx * sums_head_stride
        + block_n_idx * sums_step_stride
        + m_offs * sums_m_stride
    )
    sums_ptrs = sums_ptr + sums_offs
    tl.store(sums_ptrs, d_j, mask=m_offs < m_size)

    if N_LOAD_MASK_NEEDED:
        v_ptr_mask = block_n_offs[:, None] < n_size  # repeated otherwise triton segfault
        v = tl.load(v_ptrs + block_n_start_idx * v_k_stride, mask=v_ptr_mask, other=0.0, eviction_policy="evict_first")
    else:
        v = tl.load(v_ptrs + block_n_start_idx * v_k_stride, eviction_policy="evict_first")

    result = tl.dot(numerators.to(q_ptr.dtype.element_ty), v)

    output_offs = (
        current_batch_idx * output_batch_stride
        + current_head_idx * output_head_stride
        + block_n_idx * output_step_stride
        + (m_offs[:, None] * output_m_stride + d_range_offs[None, :] * output_n_stride)
    )

    output_ptrs = output_ptr + output_offs

    if M_LOAD_MASK_NEEDED:
        output_ptr_mask = m_offs[:, None] < m_size
        tl.store(output_ptrs, result, mask=output_ptr_mask)
    else:
        tl.store(output_ptrs, result)


@triton.jit
def _fwd_part_2(
    head_size,
    intermediates_size,
    m_size,
    input_ptr,
    input_batch_stride,
    input_head_stride,
    input_intermediate_stride,
    input_m_stride,
    input_n_stride,
    maximums_ptr,
    maximums_batch_stride,
    maximums_head_stride,
    maximums_intermediate_stride,
    maximums_m_stride,
    sums_ptr,
    sums_batch_stride,
    sums_head_stride,
    sums_intermediate_stride,
    sums_m_stride,
    output_ptr,
    output_batch_stride,
    output_head_stride,
    output_m_stride,
    output_n_stride,
    BLOCK_M_SIZE: tl.constexpr,
    BLOCK_DHEAD_SIZE: tl.constexpr,
):
    block_m_idx = tl.program_id(0)
    head_idx = tl.program_id(1)
    current_batch_idx = head_idx // head_size
    current_head_idx = head_idx % head_size

    m_range_offs = tl.arange(0, BLOCK_M_SIZE)
    dhead_range_offs = tl.arange(0, BLOCK_DHEAD_SIZE)

    m_offs = block_m_idx * BLOCK_M_SIZE + m_range_offs

    acc = tl.zeros((BLOCK_M_SIZE, BLOCK_DHEAD_SIZE), dtype=tl.float32)
    l_i = tl.zeros((BLOCK_M_SIZE,), dtype=tl.float32) - float("inf")
    d_i = tl.zeros((BLOCK_M_SIZE,), dtype=tl.float32)
    for n_intermediate_idx in range(0, intermediates_size):
        input_offs = (
            current_batch_idx * input_batch_stride
            + current_head_idx * input_head_stride
            + n_intermediate_idx * input_intermediate_stride
            + (m_offs[:, None] * input_m_stride + dhead_range_offs[None, :] * input_n_stride)
        )
        input_ptrs = input_ptr + input_offs
        numerators = tl.load(input_ptrs, mask=m_offs[:, None] < m_size, other=0.0)

        sums_offs = (
            current_batch_idx * sums_batch_stride
            + current_head_idx * sums_head_stride
            + n_intermediate_idx * sums_intermediate_stride
            + m_offs * sums_m_stride
        )
        sums_ptrs = sums_ptr + sums_offs
        d_j = tl.load(sums_ptrs, mask=m_offs < m_size, other=0.0)

        maximums_offs = (
            current_batch_idx * maximums_batch_stride
            + current_head_idx * maximums_head_stride
            + n_intermediate_idx * maximums_intermediate_stride
            + m_offs * maximums_m_stride
        )
        maximums_ptrs = maximums_ptr + maximums_offs
        l_j = tl.load(maximums_ptrs, mask=m_offs < m_size, other=0.0)

        l_new = tl.maximum(l_i, l_j)
        alpha = tl.exp(l_i - l_new)
        beta = tl.exp(l_j - l_new)
        d_new = alpha * d_i + beta * d_j

        p_scale = beta / d_new

        acc_scale = d_i / d_new * alpha
        acc *= acc_scale[:, None]

        acc += numerators * p_scale[:, None]

        d_i = d_new
        l_i = l_new

    output_offs = (
        current_batch_idx * output_batch_stride
        + current_head_idx * output_head_stride
        + (m_offs[:, None] * output_m_stride + dhead_range_offs[None, :] * output_n_stride)
    )
    output_ptrs = output_ptr + output_offs
    tl.store(output_ptrs, acc, mask=m_offs[:, None] < m_size)


class SkinnyAttention(torch.autograd.Function):
    @staticmethod
    @custom_fwd(cast_inputs=torch.float16)
    def forward(
        ctx: FunctionCtx,
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        output: torch.Tensor,
        sm_scale: float,
        is_causal: bool,
        attention_mask: Optional[torch.Tensor] = None,
    ):
        """
        Computes attention.
        FP32 input and output are not supported.
        https://github.com/openai/triton/issues/674
        Not an issue as the function is annotated with @custom_fwd(cast_inputs=torch.float16) so the input is casted to
        float16 before the function is called.

        @param ctx: context for autograd
        @param q: Query matrix size (batch, heads, size_m, dhead)
        @param k: Key matrix size (batch, heads, size_n, dhead)
        @param v: Value matrix size (batch, heads, size_n, dhead)
        @param output: Output matrix size (batch, heads, size_m, dhead)
        @param sm_scale: Scaling factor applied after operation QxK
        @param is_causal: Autoregressive decoder attention
        @param attention_mask: Attention mask matrix broadcastable to (batch, heads, size_m, size_n)
        @return:
        """
        # Constraints
        # Queries and keys have the same d_k size
        assert q.shape[-1] == k.shape[-1]
        # assert (
        #     q.dtype == k.dtype == v.dtype == output.dtype
        # ), f"All tensors must have the same dtype: {q.dtype}, {k.dtype}, {v.dtype}, {output.dtype}"
        assert q.dtype in [torch.float16, torch.bfloat16], f"Only float16 and bfloat16 are supported, got {q.dtype}"
        batch, heads, size_m, dhead = q.size()
        size_n = k.size(2)

        BLOCK_M = 16
        BLOCK_N = 128
        # load mask is needed if one dim (size_n / size_m) of tensors do not align with block size
        NEED_LOAD_MASK_SIZE_N = size_n % BLOCK_N != 0
        NEED_LOAD_MASK_SIZE_M = size_m % BLOCK_M != 0

        n_divisions = triton.cdiv(size_n, BLOCK_N)
        splitted_qkt = torch.empty(
            q.size(0), q.size(1), n_divisions, q.size(2), q.size(3), dtype=torch.float16, device="cuda"
        )

        grid = (n_divisions, triton.cdiv(size_m, BLOCK_M), batch * heads)

        maximums = torch.zeros(
            (
                batch,
                heads,
                n_divisions,
                size_m,
            ),
            device=q.device,
            dtype=torch.float32,
        )
        sums = torch.zeros(
            (
                batch,
                heads,
                n_divisions,
                size_m,
            ),
            device=q.device,
            dtype=torch.float32,
        )

        HAS_MASK = False
        if attention_mask is not None:
            assert (
                attention_mask.size(0) == batch or attention_mask.size(0) == 1
            ), "Incompatible broadcast batch dimension"
            assert (
                attention_mask.size(1) == heads or attention_mask.size(1) == 1
            ), "Incompatible broadcast heads dimension"
            assert (
                attention_mask.size(2) == size_m or attention_mask.size(2) == 1
            ), "Incompatible broadcast size_m dimension"
            assert attention_mask.size(3) == size_n, "Last size of mask must broadcast on QK^t"

            HAS_MASK = True

        _fwd_part_1[grid](
            head_size=heads,
            m_size=size_m,
            n_size=size_n,
            q_ptr=q,
            k_ptr=k,
            v_ptr=v,
            sm_scale=sm_scale,
            attention_mask_ptr=attention_mask,
            output_ptr=splitted_qkt,
            maximums_ptr=maximums,
            sums_ptr=sums,
            q_batch_stride=q.stride(0),
            q_head_stride=q.stride(1),
            q_m_stride=q.stride(2),
            q_k_stride=q.stride(3),
            k_batch_stride=k.stride(0),
            k_head_stride=k.stride(1),
            k_n_stride=k.stride(2),
            k_k_stride=k.stride(3),
            v_batch_stride=v.stride(0),
            v_head_stride=v.stride(1),
            v_k_stride=v.stride(2),
            v_n_stride=v.stride(3),
            sums_batch_stride=sums.stride(0),
            sums_head_stride=sums.stride(1),
            sums_step_stride=sums.stride(2),
            sums_m_stride=sums.stride(3),
            maximums_batch_stride=maximums.stride(0),
            maximums_head_stride=maximums.stride(1),
            maximums_step_stride=maximums.stride(2),
            maximums_m_stride=maximums.stride(3),
            output_batch_stride=splitted_qkt.stride(0),
            output_head_stride=splitted_qkt.stride(1),
            output_step_stride=splitted_qkt.stride(2),
            output_m_stride=splitted_qkt.stride(3),
            output_n_stride=splitted_qkt.stride(4),
            attention_mask_batch_stride=attention_mask.stride(0) if HAS_MASK else 0,
            attention_mask_head_stride=attention_mask.stride(1) if HAS_MASK else 0,
            attention_mask_m_stride=attention_mask.stride(2) if HAS_MASK else 0,
            attention_mask_k_stride=attention_mask.stride(3) if HAS_MASK else 0,
            N_LOAD_MASK_NEEDED=NEED_LOAD_MASK_SIZE_N,
            M_LOAD_MASK_NEEDED=NEED_LOAD_MASK_SIZE_M,
            min_clamp_value=torch.finfo(attention_mask.dtype).min if HAS_MASK else 0,
            MASK_BATCH_SIZE=attention_mask.size(0) if HAS_MASK else 0,
            MASK_HEAD_SIZE=attention_mask.size(1) if HAS_MASK else 0,
            MASK_M_SIZE=attention_mask.size(2) if HAS_MASK else 0,
            MASK_K_SIZE=attention_mask.size(3) if HAS_MASK else 0,
            HAS_MASK=HAS_MASK,
            IS_CAUSAL=is_causal,
            BLOCK_M_SIZE=BLOCK_M,
            BLOCK_N_SIZE=BLOCK_N,
            BLOCK_DHEAD_SIZE=dhead,
            num_warps=1,
            num_stages=8,
        )

        batch, heads, steps, size_m, dhead = splitted_qkt.size()
        # Attention hardcode
        BLOCK_M = 16
        grid_part2 = (triton.cdiv(size_m, BLOCK_M), batch * heads)
        _fwd_part_2[grid_part2](
            heads,
            steps,
            size_m,
            splitted_qkt,
            *splitted_qkt.stride(),
            maximums,
            *maximums.stride(),
            sums,
            *sums.stride(),
            output,
            *output.stride(),
            BLOCK_M_SIZE=BLOCK_M,
            BLOCK_DHEAD_SIZE=dhead,
            num_warps=4,
            num_stages=1,
        )
        return output


def skinny_attention_forward(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    output: torch.Tensor,
    sm_scale: float,
    is_causal: bool = False,
    attention_mask: Optional[torch.Tensor] = None,
):
    return SkinnyAttention.apply(q, k, v, output, sm_scale, is_causal, attention_mask)



import logging

import torch
import triton

from ..utils.pointwise_dynamic import pointwise_dynamic

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


@pointwise_dynamic(promotion_methods=[(0, "DEFAULT")])
@triton.jit
def conj_func(x):
    return x ^ -(1 << 63)


def resolve_conj(A: torch.Tensor):
    logger.debug("GEMS_CAMBRICON RESOLVE_CONJ")
    if A.is_conj():
        assert (
            A.dtype == torch.cfloat
        ), "The `resolve_conj` operation in FlagGems currently only supports the `torch.cfloat` type"
        typed_view = torch.view_as_real(A.conj()).view(torch.int64)
        out = conj_func(typed_view)
        return torch.view_as_complex(out.view(torch.float))
    else:
        return A



import logging
from typing import Tuple

import torch
import triton
import triton.language as tl

from flag_gems.ops.log import log as default_log  # fallback
from flag_gems.runtime import torch_device_fn
from flag_gems.utils import libentry
from flag_gems.utils.triton_lang_helper import tl_extra_shim

logger = logging.getLogger(
    f'flag_gems.runtime.backend._mthreads.ops.{__name__.split(".")[-1]}'
)

_SUPPORTED_DTYPES = {torch.float16, torch.bfloat16, torch.float32}


@libentry()
@triton.autotune(
    configs=[
        triton.Config({"BLOCK_SIZE": 256}, num_warps=4, num_stages=2),
        triton.Config({"BLOCK_SIZE": 512}, num_warps=8, num_stages=2),
        triton.Config({"BLOCK_SIZE": 1024}, num_warps=8, num_stages=2),
        triton.Config({"BLOCK_SIZE": 2048}, num_warps=8, num_stages=2),
        triton.Config({"BLOCK_SIZE": 4096}, num_warps=16, num_stages=2),
    ],
    key=["n_elements", "dtype_size"],
)
@triton.jit
def log_kernel(
    x_ptr,
    out_ptr,
    n_elements,
    dtype_size,
    BLOCK_SIZE: tl.constexpr,
    USE_APPROX: tl.constexpr,
):
    pid = tl.program_id(0)
    offsets = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    mask = offsets < n_elements
    x = tl.load(x_ptr + offsets, mask=mask, other=0.0)
    x_fp32 = x.to(tl.float32)
    if USE_APPROX:
        pos_mask = x_fp32 > 0
        zero_mask = x_fp32 == 0
        ix = x_fp32.to(tl.int32, bitcast=True)
        exp = (ix >> 23) & 0xFF
        mant = (ix & 0x7FFFFF) | 0x3F800000
        m = mant.to(tl.float32, bitcast=True)
        k = exp.to(tl.int32) - 127
        t = (m - 1.0) / (m + 1.0)
        t2 = t * t
        # t4 = t2 * t2
        # t6 = t4 * t2
        log_m = 2.0 * (t + t2 * t * (1.0 / 3.0 + t2 * (1.0 / 5.0 + t2 * (1.0 / 7.0))))
        log_val = log_m + k.to(tl.float32) * 0.6931471805599453
        nan_or_inf = tl.where(zero_mask, -float("inf"), float("nan"))
        y = tl.where(pos_mask, log_val, nan_or_inf)
    else:
        y = tl_extra_shim.log(x_fp32)
    tl.store(out_ptr + offsets, y, mask=mask)


def _use_triton_kernel(x: torch.Tensor) -> Tuple[bool, int]:
    if not isinstance(x, torch.Tensor):
        return False, 0
    if x.device.type != "musa" or x.dtype not in _SUPPORTED_DTYPES:
        return False, 0
    if x.numel() == 0 or not x.is_contiguous():
        return False, 0
    return True, x.element_size()


def _launch_log(x: torch.Tensor, out: torch.Tensor, dtype_size: int):
    n_elements = out.numel()
    grid = lambda meta: (triton.cdiv(n_elements, meta["BLOCK_SIZE"]),)
    with torch_device_fn.device(out.device):
        log_kernel[grid](x, out, n_elements, dtype_size, USE_APPROX=dtype_size == 2)
    return out


def log(x):
    logger.debug("GEMS_MTHREADS LOG")
    use_triton, dtype_size = _use_triton_kernel(x)
    if not use_triton:
        return default_log(x)

    out = torch.empty_like(x)
    return _launch_log(x, out, dtype_size)



import logging

import triton
import triton.language as tl

from flag_gems.utils import pointwise_dynamic

logger = logging.getLogger(__name__)


@pointwise_dynamic(promotion_methods=[(0, "DEFAULT")])
@triton.jit
def relu_forward(x):
    return tl.where(x > 0, x, 0)


@pointwise_dynamic(promotion_methods=[(0, "DEFAULT")])
@triton.jit
def relu_backward(x, dy):
    return tl.where(x > 0, dy, 0)


def relu(self):
    logger.debug("GEMS RELU FORWARD")
    output = relu_forward(self)
    return output


def relu_(A):
    logger.debug("GEMS RELU_ FORWARD")
    out = relu_forward(A, out0=A)
    return out



#  Copyright (c) NXAI GmbH.
#  This software may be used and distributed according to the terms of the NXAI Community License Agreement.

# Maximilian Beck
"""
Triton.

Backward parallel kernel of the mLSTM chunkwise formulation.

Notation:
Dimensions:
    B: batch size
    NH: number of heads
    S: sequence length (K, V)
    T: sequence length (Q)
    DHQK: hidden dimension (Q, K)
    DHHV: hidden dimension (H, V)
    NC: number of chunks
    L: chunk size

Variables:
    vecA, a: forward gate contribution, contribution of forget gates from last chunk state C_{k-1} to current timestep t
    vecB, b: backward gate contribution, contribution of forget and input gates up to next chunk state C_k (form current timestep t)
    scaG, g: "go through" gate contribution, contribution of forget gates from C_{k-1} to C_k.
    matD, D: gating matrix for the parallel form.
"""

import triton
import triton.language as tl


@triton.jit
def mlstm_chunkwise__parallel_bw_dQKV_kernel(
    matQ,  # (B, NH, S, DHQK)
    matK,  # (B, NH, S, DHQK)
    matV,  # (B, NH, S, DHHV)
    vecB,  # (B, NH, NC, L)
    vecI,  # (B, NH, NC, L)
    vecM_combine,  # (B, NH, S)
    scaM_inter,  # (B, NH, NC+1)
    matC_states,  # (B, NH, (NC+1) * DHQK, DHHV) # take only the first NC states
    matDeltaH,  # (B, NH, S, DHHV)
    vecN_out,  # (B, NH, S)
    matDeltaC_states,  # (B, NH, (NC+1) * DHQK, DHHV) # take only the last NC states
    matDeltaQ,  # (B, NH, S, DHQK)
    matDeltaK,  # (B, NH, S, DHQK)
    matDeltaV,  # (num_b_DHQK, B, NH, S, DHHV)
    qk_scale,
    str_matQK_B_NH,  # shared with matQ, matDeltaQ, matK, matDeltaK
    str_matQK_S,
    str_matQK_DHQK,
    str_matDV_num_b_DHQK,
    str_matHV_B_NH,  # shared with matDeltaV, matDeltaH
    str_matHV_S,
    str_matHV_DHHV,
    str_vecBI_B_NH,
    str_vecBI_NC,
    str_vecBI_L,
    str_vecM_combine_B_NH,
    str_vecM_combine_S,
    str_scaM_inter_B_NH,
    str_scaM_inter_NC,
    str_matC_states_B_NH,
    str_matC_states_NCDHQK,
    str_matC_states_DHHV,
    str_vecN_out_B_NH,
    str_vecN_out_S,
    str_matDeltaC_states_B_NH,
    str_matDeltaC_states_NCDHQK,
    str_matDeltaC_states_DHHV,
    B: tl.constexpr,
    NH: tl.constexpr,
    S: tl.constexpr,
    DHQK: tl.constexpr,
    DHHV: tl.constexpr,
    NC: tl.constexpr,
    L: tl.constexpr,
    siz_b_DHQK: tl.constexpr,
    siz_b_DHHV: tl.constexpr,
    DTYPE: tl.constexpr = tl.float32,
    EPS: tl.constexpr = 1e-6,
):
    idx_b_DHQK, idx_b_NC, idx_b_BNH = (
        tl.program_id(0),
        tl.program_id(1),
        tl.program_id(2),
    )

    # [intra] recompute matDbar
    # load gates (L,)
    vecB_val = tl.load(vecB + idx_b_BNH * S + idx_b_NC * L + tl.arange(0, L)).to(
        tl.float32
    )  # (L,)
    vecI_val = tl.load(vecI + idx_b_BNH * S + idx_b_NC * L + tl.arange(0, L)).to(
        tl.float32
    )  # (L,)
    # load vecM_combine (L,)
    vecM_combine_val = tl.load(
        vecM_combine + idx_b_BNH * S + idx_b_NC * L + tl.arange(0, L)
    )

    # compute gate matrix matDbar (L, L)
    idx_mask = tl.arange(0, L)
    mask = idx_mask[:, None] >= idx_mask[None, :]
    matF_full_val = vecB_val[:, None] - vecB_val[None, :]
    matF_mask_val = tl.where(mask, matF_full_val, -float("inf"))
    matDtilde_val = matF_mask_val + vecI_val[None, :]
    matDbar_val = tl.exp(matDtilde_val - vecM_combine_val[:, None])

    # [inter,intra] compute vecAbar, vecBbar
    # load scaM_inter_k, scaM_inter_km1
    scaM_inter_km1_val = tl.load(scaM_inter + idx_b_BNH * (NC + 1) + idx_b_NC)  # (1,)
    scaM_inter_k_val = tl.load(
        scaM_inter + idx_b_BNH * (NC + 1) + (idx_b_NC + 1)
    )  # (1,)

    # scaG is the last val of vecB
    scaG_val = tl.load(vecB + idx_b_BNH * S + idx_b_NC * L + L - 1)  # (1,)
    vecA_val = scaG_val - vecB_val + vecI_val  # (L,)

    vecAbar_val = tl.exp(vecA_val - scaM_inter_k_val)  # (L,)
    vecBbar_val = tl.exp(vecB_val + scaM_inter_km1_val - vecM_combine_val)  # (L, )

    # [intra] recompute matS, matSbar
    # NOTE: we compute only a part of matS as we do not sum over the full DHQK dim
    # we have to sum the matV after the kernel.
    matQ_ptr = tl.make_block_ptr(
        base=matQ + idx_b_BNH * str_matQK_B_NH,
        shape=(S, DHQK),
        strides=(str_matQK_S, str_matQK_DHQK),
        offsets=(idx_b_NC * L, idx_b_DHQK * siz_b_DHQK),
        block_shape=(L, siz_b_DHQK),
        order=(1, 0),
    )
    matK_ptr = tl.make_block_ptr(
        base=matK + idx_b_BNH * str_matQK_B_NH,
        shape=(S, DHQK),
        strides=(str_matQK_S, str_matQK_DHQK),
        offsets=(idx_b_NC * L, idx_b_DHQK * siz_b_DHQK),
        block_shape=(L, siz_b_DHQK),
        order=(1, 0),
    )
    matQ_val = tl.load(matQ_ptr, boundary_check=(0, 1)).to(DTYPE)  # (L, siz_b_DHQK)
    matK_val = tl.load(matK_ptr, boundary_check=(0, 1)).to(DTYPE)  # (siz_b_DHQK, L)
    matS_val = tl.dot(matQ_val, tl.trans(matK_val)) * qk_scale
    matSbar_val = (matS_val * matDbar_val).to(DTYPE)  # (L, L)

    matKbar_val = (matK_val.to(tl.float32) * vecAbar_val[:, None]).to(
        DTYPE
    )  # (L, siz_b_DHQK)

    vecN_out_ptr = vecN_out + idx_b_BNH * S + idx_b_NC * L + tl.arange(0, L)
    vecN_out_val = tl.load(vecN_out_ptr).to(tl.float32)  # (L,)
    matDeltaQ_inter_val = tl.zeros((L, siz_b_DHQK), dtype=tl.float32)
    matDeltaK_inter_val = tl.zeros((L, siz_b_DHQK), dtype=tl.float32)
    matDeltaSbar_val = tl.zeros((L, L), dtype=tl.float32)

    for idx_b_DHHV in range(tl.cdiv(DHHV, siz_b_DHHV)):
        # ? pointers for iteration
        # load pointers:
        matV_ptr = tl.make_block_ptr(
            base=matV + idx_b_BNH * str_matHV_B_NH,
            shape=(S, DHHV),
            strides=(str_matHV_S, str_matHV_DHHV),
            offsets=(idx_b_NC * L, idx_b_DHHV * siz_b_DHHV),
            block_shape=(L, siz_b_DHHV),
            order=(1, 0),
        )
        matDeltaH_ptr = tl.make_block_ptr(
            base=matDeltaH + idx_b_BNH * str_matHV_B_NH,
            shape=(S, DHHV),
            strides=(str_matHV_S, str_matHV_DHHV),
            offsets=(idx_b_NC * L, idx_b_DHHV * siz_b_DHHV),
            block_shape=(L, siz_b_DHHV),
            order=(1, 0),
        )
        matC_km1_ptr = tl.make_block_ptr(
            base=matC_states
            + idx_b_BNH * str_matC_states_B_NH
            + idx_b_NC * DHQK * DHHV,
            shape=(DHHV, DHQK),  # transposed
            strides=(str_matC_states_DHHV, str_matC_states_NCDHQK),
            offsets=(idx_b_DHHV * siz_b_DHHV, idx_b_DHQK * siz_b_DHQK),
            block_shape=(siz_b_DHHV, siz_b_DHQK),
            order=(0, 1),
        )
        # (idx_b_NC + 1) since we take only the last NC states
        matDeltaC_k_ptr = tl.make_block_ptr(
            base=matDeltaC_states
            + idx_b_BNH * str_matDeltaC_states_B_NH
            + (idx_b_NC + 1) * DHQK * DHHV,
            shape=(DHQK, DHHV),
            strides=(str_matDeltaC_states_NCDHQK, str_matDeltaC_states_DHHV),
            offsets=(idx_b_DHQK * siz_b_DHQK, idx_b_DHHV * siz_b_DHHV),
            block_shape=(siz_b_DHQK, siz_b_DHHV),
            order=(1, 0),
        )
        # store pointers:
        matDeltaV_ptr = tl.make_block_ptr(
            base=matDeltaV
            + idx_b_DHQK * str_matDV_num_b_DHQK
            + idx_b_BNH * str_matHV_B_NH,
            shape=(S, DHHV),
            strides=(str_matHV_S, str_matHV_DHHV),
            offsets=(idx_b_NC * L, idx_b_DHHV * siz_b_DHHV),
            block_shape=(L, siz_b_DHHV),
            order=(1, 0),
        )
        # ? end pointers

        # [inter,intra] matDeltaV
        # matDeltaH = matDeltaH / vecN_out
        matDeltaH_val = tl.load(matDeltaH_ptr, boundary_check=(0, 1)).to(
            tl.float32
        )  # (L, siz_b_DHHV)
        # matDeltaH_val = matDeltaH_val / (vecN_out_val[:, None] + EPS)  # (L, siz_b_DHHV)

        # [inter] matDeltaK += (matV @ matDeltaC_k.transpose()) * vecAbar
        matV_val = tl.load(matV_ptr, boundary_check=(0, 1)).to(DTYPE)  # (L, siz_b_DHHV)
        matDeltaC_k_val = tl.load(matDeltaC_k_ptr, boundary_check=(0, 1)).to(
            DTYPE
        )  # (siz_b_DHQK, siz_b_DHHV)

        matDeltaK_inter_val += tl.dot(
            matV_val, tl.trans(matDeltaC_k_val)
        )  # (L, siz_b_DHQK)

        # [inter] matDeltaQ += matDeltaH @ matC_km1.transpose() * vecBbar
        matC_km1_val = tl.load(matC_km1_ptr, boundary_check=(0, 1)).to(
            DTYPE
        )  # (siz_b_DHHV, siz_b_DHQK)
        matDeltaQ_inter_val += (
            tl.dot((matDeltaH_val).to(DTYPE), matC_km1_val) * qk_scale
        )  # (L, siz_b_DHQK)

        # [intra] matDeltaS += (matDeltaH @ matV.transpose()) * matDbar
        matDeltaSbar_val += tl.dot(
            (matDeltaH_val / (vecN_out_val[:, None] + EPS)).to(DTYPE),
            tl.trans(matV_val),
        )  # (L, L)

        # [inter, intra] matDeltaV = matSbar.transpose() @ matDeltaH + matKbar @ matDeltaC_k
        matDeltaV_val = tl.dot(
            tl.trans(matSbar_val),
            (matDeltaH_val / (vecN_out_val[:, None] + EPS)).to(DTYPE),
        ) + tl.dot(matKbar_val, matDeltaC_k_val)
        # store matDeltaV (NOTE: each idx_b_DHQK stores its own contribution in HBM, it will be summed outside the kernel)
        tl.store(
            matDeltaV_ptr,
            matDeltaV_val.to(matDeltaV_ptr.dtype.element_ty),
            boundary_check=(0, 1),
        )
    matDeltaQ_inter_val = (
        matDeltaQ_inter_val * vecBbar_val[:, None] / (vecN_out_val[:, None] + EPS)
    )
    matDeltaK_inter_val = matDeltaK_inter_val * vecAbar_val[:, None]

    matDeltaS_val = (matDeltaSbar_val * matDbar_val).to(DTYPE)

    # [intra] matDeltaK = matDeltaS.transpose() @ matQ * qk_scale
    matDeltaK_val = matDeltaK_inter_val + (
        tl.dot(tl.trans(matDeltaS_val), (matQ_val).to(DTYPE)) * qk_scale
    )
    # [intra] matDeltaQ = matDeltaS @ matK * qk_scale
    matDeltaQ_val = matDeltaQ_inter_val + ((tl.dot(matDeltaS_val, matK_val)) * qk_scale)

    # store matDeltaQ, matDeltaK
    matDeltaK_ptr = tl.make_block_ptr(
        base=matDeltaK + idx_b_BNH * str_matQK_B_NH,
        shape=(S, DHQK),
        strides=(str_matQK_S, str_matQK_DHQK),
        offsets=(idx_b_NC * L, idx_b_DHQK * siz_b_DHQK),
        block_shape=(L, siz_b_DHQK),
        order=(1, 0),
    )
    matDeltaQ_ptr = tl.make_block_ptr(
        base=matDeltaQ + idx_b_BNH * str_matQK_B_NH,
        shape=(S, DHQK),
        strides=(str_matQK_S, str_matQK_DHQK),
        offsets=(idx_b_NC * L, idx_b_DHQK * siz_b_DHQK),
        block_shape=(L, siz_b_DHQK),
        order=(1, 0),
    )
    tl.store(
        matDeltaK_ptr,
        matDeltaK_val.to(matDeltaK.dtype.element_ty),
        boundary_check=(0, 1),
    )
    tl.store(
        matDeltaQ_ptr,
        matDeltaQ_val.to(matDeltaQ.dtype.element_ty),
        boundary_check=(0, 1),
    )



#  Copyright (c) NXAI GmbH.
#  This software may be used and distributed according to the terms of the NXAI Community License Agreement.

# Maximilian Beck
"""
Triton.

This module contains the recurrent step of the mLSTM in triton.

This is a fused forward decoding step kernel for the mLSTM. Factor of 2 speedup compared to torch.compile.
Ca. 30% faster than non-fused triton step kernel.
"""

import triton
import triton.language as tl


@triton.jit
def recurrent_step_fw_kernel(
    matC_old,  # (B, NH, DHQK, DHHV)
    vecN_old,  # (B, NH, DHQK)
    scaM_old,  # (B, NH, 1)
    vecQ,  # (B, NH, DHQK)
    vecK,  # (B, NH, DHQK)
    vecV,  # (B, NH, DHHV)
    scaI,  # (B, NH, 1)
    scaF,  # (B, NH, 1)
    vecH,  # (B, NH, DHHV)
    matC_new,  # (B, NH, DHQK, DHHV)
    vecN_new,  # (B, NH, DHQK)
    scaM_new,  # (B, NH, 1)
    qk_scale: tl.constexpr,
    str_matC_B_NH: tl.constexpr,
    str_matC_DHQK: tl.constexpr,
    str_matC_DHHV: tl.constexpr,
    str_vecN_B_NH: tl.constexpr,
    str_vecN_DHQK: tl.constexpr,
    str_scaM_B_NH: tl.constexpr,
    str_vecQK_NH: tl.constexpr,
    str_vecQK_DHQK: tl.constexpr,
    str_vecVH_B_NH: tl.constexpr,
    str_vecVH_DHHV: tl.constexpr,
    str_scaIF_B_NH: tl.constexpr,
    B: tl.constexpr,
    NH: tl.constexpr,
    DHQK: tl.constexpr,
    DHHV: tl.constexpr,
    siz_b_DHQK: tl.constexpr,
    siz_b_DHHV: tl.constexpr,
    EPS: tl.constexpr = 1e-6,
    DTYPE: tl.constexpr = tl.float32,
    DTYPE_STATE: tl.constexpr = tl.float32,
):
    idx_b_DHHV, idx_b_BNH = tl.program_id(1), tl.program_id(2)

    # ? Define pointers
    matC_old_bptr = tl.make_block_ptr(
        base=matC_old + idx_b_BNH * str_matC_B_NH,
        shape=(DHQK, DHHV),
        strides=(str_matC_DHQK, str_matC_DHHV),
        offsets=(0, idx_b_DHHV * siz_b_DHHV),
        block_shape=(siz_b_DHQK, siz_b_DHHV),
        order=(0, 1),
    )
    matC_new_bptr = tl.make_block_ptr(
        base=matC_new + idx_b_BNH * str_matC_B_NH,
        shape=(DHQK, DHHV),
        strides=(str_matC_DHQK, str_matC_DHHV),
        offsets=(0, idx_b_DHHV * siz_b_DHHV),
        block_shape=(siz_b_DHQK, siz_b_DHHV),
        order=(0, 1),
    )
    vecH_ptr = (
        vecH
        + idx_b_BNH * str_vecVH_B_NH
        + idx_b_DHHV * siz_b_DHHV * str_vecVH_DHHV
        + tl.arange(0, siz_b_DHHV)
    )

    scaI_ptr = scaI + idx_b_BNH * str_scaIF_B_NH
    scaF_ptr = scaF + idx_b_BNH * str_scaIF_B_NH

    scaM_old_ptr = scaM_old + idx_b_BNH * str_scaM_B_NH
    scaM_new_ptr = scaM_new + idx_b_BNH * str_scaM_B_NH

    # ? Load data
    # gates
    # the numbers are the conversion factors from log -> log2 and exp -> exp2
    # math.log2(math.e) = 1.4426950408889634
    # (1/math.log2(math.e)) = 0.6931471805599453
    # tl.exp and tl.sigmoid only work with float32
    scaF_val = tl.load(scaF_ptr).to(tl.float32)
    scaI_val = tl.load(scaI_ptr).to(tl.float32)
    scaFlog_val = tl.log(tl.sigmoid(scaF_val))

    scaM_old_val = tl.load(scaM_old_ptr)
    scaM_new_val = tl.maximum(scaFlog_val + scaM_old_val, scaI_val)
    if idx_b_DHHV == 0:  # only one thread block writes the scaM_new
        tl.store(scaM_new_ptr, scaM_new_val.to(DTYPE_STATE))

    max_val = tl.exp(-scaM_new_val.to(tl.float32)).to(DTYPE)

    # gate computation for all dimensions
    scaF_act = tl.exp(scaFlog_val + scaM_old_val - scaM_new_val).to(DTYPE)
    scaI_act = tl.exp(scaI_val - scaM_new_val).to(DTYPE)
    # tl.static_print("scaF_act", scaF_act)
    # ? init accumulators
    h_num = tl.zeros((siz_b_DHHV,), dtype=tl.float32)
    qn_dotproduct = tl.zeros((1,), dtype=tl.float32)

    NUM_BLOCKS_DQK = triton.cdiv(DHQK, siz_b_DHQK)

    for i_dhqk in range(NUM_BLOCKS_DQK):
        vecN_old_ptr = (
            vecN_old
            + idx_b_BNH * str_vecN_B_NH
            + i_dhqk * siz_b_DHQK * str_vecN_DHQK
            + tl.arange(0, siz_b_DHQK)
        )
        vecN_new_ptr = (
            vecN_new
            + idx_b_BNH * str_vecN_B_NH
            + i_dhqk * siz_b_DHQK * str_vecN_DHQK
            + tl.arange(0, siz_b_DHQK)
        )

        vecQ_ptr = (
            vecQ
            + idx_b_BNH * str_vecQK_NH
            + i_dhqk * siz_b_DHQK * str_vecQK_DHQK
            + tl.arange(0, siz_b_DHQK)
        )
        vecK_ptr = (
            vecK
            + idx_b_BNH * str_vecQK_NH
            + i_dhqk * siz_b_DHQK * str_vecQK_DHQK
            + tl.arange(0, siz_b_DHQK)
        )
        vecV_ptr = (
            vecV
            + idx_b_BNH * str_vecVH_B_NH
            + idx_b_DHHV * siz_b_DHHV * str_vecVH_DHHV
            + tl.arange(0, siz_b_DHHV)
        )

        # update rule
        vecK_val = tl.load(vecK_ptr)
        vecV_val = tl.load(vecV_ptr)
        matC_old_val = tl.load(
            matC_old_bptr, boundary_check=(0, 1), padding_option="zero"
        ).to(dtype=DTYPE_STATE)
        matC_old_val = tl.load(
            matC_old_bptr, boundary_check=(0, 1), padding_option="zero"
        ).to(dtype=DTYPE_STATE)

        matC_new_val = scaF_act * matC_old_val + scaI_act * (
            vecK_val[:, None] * vecV_val[None, :]
        )
        matC_new_val = scaF_act * matC_old_val + scaI_act * (
            vecK_val[:, None] * vecV_val[None, :]
        )

        vecN_new_val = (
            scaF_act * tl.load(vecN_old_ptr).to(dtype=DTYPE_STATE) + scaI_act * vecK_val
        )
        vecN_new_val = (
            scaF_act * tl.load(vecN_old_ptr).to(dtype=DTYPE_STATE) + scaI_act * vecK_val
        )
        # ? Store data
        tl.store(
            matC_new_bptr,
            matC_new_val.to(matC_new.type.element_ty),
            boundary_check=(0, 1),
        )
        if idx_b_DHHV == 0:  # only one thread block writes the vecN_new
            tl.store(vecN_new_ptr, vecN_new_val.to(vecN_new.type.element_ty))

        # ? advance pointers
        matC_old_bptr = tl.advance(matC_old_bptr, (siz_b_DHQK, 0))
        matC_new_bptr = tl.advance(matC_new_bptr, (siz_b_DHQK, 0))

        # ? accumulate h_num & qn_dotproduct
        vecQ_val = tl.load(vecQ_ptr) * qk_scale
        # outputs
        h_num_temp = vecQ_val[:, None] * matC_new_val.to(dtype=DTYPE)
        # we keep h_num and qn_dotproduct in float32 as they are accumulated
        h_num += tl.sum(h_num_temp, axis=0)
        qn_dotproduct += tl.sum(vecQ_val * vecN_new_val.to(dtype=DTYPE))

    # we compute h in float32 and then cast to DTYPE
    h_denom = tl.maximum(tl.abs(qn_dotproduct), max_val) + EPS
    h = tl.fdiv(h_num, h_denom)

    # ? Store data
    tl.store(vecH_ptr, h.to(DTYPE))



import logging

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.runtime import device, torch_device_fn
from flag_gems.utils.random_utils import (
    philox_backend_seed_offset,
    uint_to_uniform_float,
)
from flag_gems.utils.shape_utils import volume

try:
    pair_uniform_to_normal = tl.pair_uniform_to_normal
except AttributeError:

    @triton.jit
    def pair_uniform_to_normal(u1, u2):
        """Box-Muller transform"""
        u1 = tl.maximum(1.0e-7, u1)
        th = 6.283185307179586 * u2
        r = tl.sqrt(-2.0 * tl.log(u1))
        return r * tl.cos(th), r * tl.sin(th)


device_ = device
logger = logging.getLogger(__name__)


@triton.heuristics(runtime.get_heuristic_config("randn"))
@triton.jit(do_not_specialize=["philox_seed", "philox_offset"])
def randn_kernel(
    out_ptr,
    N,
    philox_seed,
    philox_offset,
    BLOCK: tl.constexpr,
):
    philox_seed = philox_seed.to(tl.int64)
    philox_offset = philox_offset.to(tl.int64)
    c0 = (philox_offset & 0xFFFFFFFF).to(tl.uint32)
    c1 = ((philox_offset >> 32) & 0xFFFFFFFF).to(tl.uint32)
    i4 = tl.program_id(0) * BLOCK + tl.arange(0, BLOCK)
    c0 += i4
    _O = c0 * 0
    r0, r1, r2, r3 = tl.philox(philox_seed, c0, c1, _O, _O)
    r0 = uint_to_uniform_float(r0)
    r1 = uint_to_uniform_float(r1)
    r2 = uint_to_uniform_float(r2)
    r3 = uint_to_uniform_float(r3)
    n0, n1 = pair_uniform_to_normal(r0, r1)
    n2, n3 = pair_uniform_to_normal(r2, r3)
    off_0 = tl.program_id(0) * BLOCK * 4 + tl.arange(0, BLOCK)
    off_1 = off_0 + BLOCK
    off_2 = off_1 + BLOCK
    off_3 = off_2 + BLOCK
    tl.store(out_ptr + off_0, n0, mask=off_0 < N, eviction_policy="evict_first")
    tl.store(out_ptr + off_1, n1, mask=off_1 < N, eviction_policy="evict_first")
    tl.store(out_ptr + off_2, n2, mask=off_2 < N, eviction_policy="evict_first")
    tl.store(out_ptr + off_3, n3, mask=off_3 < N, eviction_policy="evict_first")


UNROLL = 4


def randn(size, *, dtype=None, layout=None, device=None, pin_memory=None):
    logger.debug("GEMS RANDN")
    if dtype is None:
        dtype = torch.get_default_dtype()
    if device is None:
        device = torch.device(device_.name)
    out = torch.empty(size, device=device, dtype=dtype)
    N = volume(size)
    grid_fn = lambda meta: (triton.cdiv(N, meta["BLOCK"] * UNROLL),)
    # (TODO) Using Triton autotuner makes kernel parameters opaque to the caller,
    # hence we cannot obtain the per thread offset as in Pytorch.
    increment = triton.cdiv(N, UNROLL)
    philox_seed, philox_offset = philox_backend_seed_offset(increment)
    with torch_device_fn.device(device):
        randn_kernel[grid_fn](out, N, philox_seed, philox_offset)
    return out



"""
Block Detection Utilities

Shared utilities for detecting block structure from weight names.
Used by Pipeline, TensorParallel, and Zero3 strategies.

ZERO HARDCODE: Pattern-based detection from weight keys.
"""

import re
from typing import Dict, List, Optional, Set


# Common block naming patterns across model architectures
BLOCK_PATTERNS = [
    r'blocks?[._](\d+)',              # blocks.0, block_0, block.1
    r'layers?[._](\d+)',              # layer.0, layers_0, layer.1
    r'transformer[._]h[._](\d+)',     # transformer.h.0 (GPT style)
    r'encoder[._]layer[._](\d+)',     # encoder.layer.0 (BERT style)
    r'decoder[._]layer[._](\d+)',     # decoder.layer.0
    r'model[._]layers[._](\d+)',      # model.layers.0 (LLaMA style)
    r'resblocks[._](\d+)',            # resblocks.0 (CLIP style)
    r'down_blocks[._](\d+)',          # down_blocks.0 (UNet style)
    r'up_blocks[._](\d+)',            # up_blocks.0 (UNet style)
    r'mid_block[._](\d+)',            # mid_block.0 (UNet style)
    r'joint_blocks[._](\d+)',         # joint_blocks.0 (DiT style)
    r'single_transformer_blocks[._](\d+)',  # Flux style
    r'transformer_blocks[._](\d+)',   # transformer_blocks.0
]


def extract_block_index(weight_name: str) -> Optional[int]:
    """
    Extract block index from a weight name.

    Args:
        weight_name: Full weight key (e.g., "transformer.blocks.5.attn.qkv.weight")

    Returns:
        Block index (e.g., 5) or None if no block pattern found

    Example:
        >>> extract_block_index("transformer.blocks.5.attn.qkv.weight")
        5
        >>> extract_block_index("embed_tokens.weight")
        None
    """
    for pattern in BLOCK_PATTERNS:
        match = re.search(pattern, weight_name, re.IGNORECASE)
        if match:
            return int(match.group(1))
    return None


def extract_block_indices_from_paths(paths: List[str]) -> List[int]:
    """
    Extract all unique block indices from a list of weight paths.

    Args:
        paths: List of weight paths/keys

    Returns:
        Sorted list of unique block indices found

    Example:
        >>> paths = ["blocks.0.attn.weight", "blocks.1.attn.weight", "embed.weight"]
        >>> extract_block_indices_from_paths(paths)
        [0, 1]
    """
    indices: Set[int] = set()

    for path in paths:
        idx = extract_block_index(path)
        if idx is not None:
            indices.add(idx)

    return sorted(indices)


def parse_block_structure(
    weight_keys: List[str],
    include_non_block: bool = True,
) -> Dict[int, List[str]]:
    """
    Parse weight keys into block-organized structure.

    Groups weights by their block index. Non-block weights (embeddings,
    final layers) are optionally assigned to block -1.

    Args:
        weight_keys: List of all weight keys
        include_non_block: If True, non-block weights go to key -1

    Returns:
        Dict mapping block_index -> list of weight keys
        Block -1 contains non-block weights if include_non_block=True

    Example:
        >>> keys = ["blocks.0.attn.weight", "blocks.1.attn.weight", "embed.weight"]
        >>> structure = parse_block_structure(keys)
        >>> structure[0]
        ['blocks.0.attn.weight']
        >>> structure[-1]  # Non-block weights
        ['embed.weight']
    """
    block_weights: Dict[int, List[str]] = {}
    non_block_weights: List[str] = []

    for key in weight_keys:
        block_idx = extract_block_index(key)

        if block_idx is not None:
            if block_idx not in block_weights:
                block_weights[block_idx] = []
            block_weights[block_idx].append(key)
        else:
            non_block_weights.append(key)

    # Add non-block weights to special key -1
    if include_non_block and non_block_weights:
        block_weights[-1] = non_block_weights

    return block_weights


def classify_non_block_weight(weight_name: str) -> str:
    """
    Classify a non-block weight as 'early' or 'late' in the model.

    Early weights (embeddings, input layers) should be processed first.
    Late weights (output layers, final norms) should be processed last.

    Args:
        weight_name: Weight key without block index

    Returns:
        'early' or 'late'
    """
    name_lower = weight_name.lower()

    # Early layers (input side)
    early_hints = [
        'embed', 'patch', 'pos', 'input', 'norm_pre', 'time_',
        'conv_in', 'proj_in', 'x_embedder', 't_embedder',
        'text_proj', 'context_embedder', 'adaln_single',
    ]

    for hint in early_hints:
        if hint in name_lower:
            return 'early'

    # Everything else is late (output side)
    return 'late'


def get_block_count(weight_keys: List[str]) -> int:
    """
    Get the number of unique blocks in weight keys.

    Args:
        weight_keys: List of weight keys

    Returns:
        Number of unique blocks (excluding non-block weights)
    """
    indices = extract_block_indices_from_paths(weight_keys)
    return len(indices)



# Copyright (c) 2023-2025, Songlin Yang, Yu Zhang


import torch
import triton
import triton.language as tl

from fla.ops.utils import prepare_chunk_indices
from fla.ops.utils.op import exp


@triton.heuristics({
    'IS_VARLEN': lambda args: args['cu_seqlens'] is not None,
})
@triton.jit(do_not_specialize=['T'])
def chunk_mesa_net_h_kk_bwd_intra_kernel(
    k,
    beta,
    h,
    dh,
    g,
    q_star,
    dq,
    dk,
    dg,
    dbeta,
    dk_beta,
    dlamb,
    cu_seqlens,
    chunk_indices,
    B: tl.constexpr,
    T,
    H: tl.constexpr,
    K: tl.constexpr,
    V: tl.constexpr,
    BT: tl.constexpr,
    BK: tl.constexpr,
    BV: tl.constexpr,
    IS_VARLEN: tl.constexpr,
):
    i_t, i_bh = tl.program_id(0), tl.program_id(1)
    i_b, i_h = i_bh // H, i_bh % H

    if IS_VARLEN:
        i_tg = i_t
        i_n, i_t = tl.load(chunk_indices + i_t * 2).to(tl.int32), tl.load(chunk_indices + i_t * 2 + 1).to(tl.int32)
        bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(cu_seqlens + i_n + 1).to(tl.int32)
        T = eos - bos
        NT = tl.cdiv(T, BT)
    else:
        NT = tl.cdiv(T, BT)
        i_tg = i_b * NT + i_t
        bos, eos = i_b * T, i_b * T + T

    o_t = i_t * BT + tl.arange(0, BT)
    m_t = o_t < T

    # offset calculation
    q_star += (bos * H + i_h) * V
    dq += (bos * H + i_h) * V
    h += (i_tg * H + i_h).to(tl.int64) * K*V
    dh += (i_tg * H + i_h).to(tl.int64) * K*V
    k += (bos * H + i_h) * K
    dk += (bos * H + i_h) * K
    dk_beta += (bos * H + i_h) * K
    dlamb += (i_tg * H + i_h).to(tl.int64) * K
    beta += (bos * H + i_h)
    dbeta += (bos * H + i_h)
    g += bos * H + i_h
    dg += bos * H + i_h

    b_dk = tl.zeros([BT, BK], dtype=tl.float32)
    b_dv = tl.zeros([BT, BK], dtype=tl.float32)
    b_dbeta = tl.zeros([BT ], dtype=tl.float32)
    b_dg_last = tl.zeros([1], dtype=tl.float32)
    b_dg = tl.zeros([BT], dtype=tl.float32)

    p_g = tl.make_block_ptr(g, (T,), (H,), (i_t * BT,), (BT,), (0,))
    b_g = tl.load(p_g, boundary_check=(0,))
    b_g_last = tl.load(g + (min(i_t * BT + BT, T) - 1) * H)
    b_gk = tl.where(m_t, exp(b_g_last - b_g), 0)

    p_q_star = tl.make_block_ptr(q_star, (T, V), (H*V, 1), (i_t * BT, 0), (BT, BV), (1, 0))
    b_q_star = tl.load(p_q_star, boundary_check=(0, 1))
    p_dq = tl.make_block_ptr(dq, (T, V), (H*V, 1), (i_t * BT, 0), (BT, BV), (1, 0))
    b_dq = tl.load(p_dq, boundary_check=(0, 1))
    b_dlamb = -tl.sum(b_q_star * b_dq, axis=0)
    p_dlamb = tl.make_block_ptr(dlamb, (K,), (1,), (0,), (BK,), (0,))
    tl.store(p_dlamb, b_dlamb.to(p_dlamb.dtype.element_ty), boundary_check=(0,))

    p_h = tl.make_block_ptr(h, (V, K), (1, V), (0, 0), (BV, BK), (0, 1))
    p_dh = tl.make_block_ptr(dh, (V, K), (1, V), (0, 0), (BV, BK), (0, 1))
    p_beta = tl.make_block_ptr(beta, (T,), (H,), (i_t * BT,), (BT,), (0,))
    b_beta = tl.load(p_beta, boundary_check=(0,))
    p_k = tl.make_block_ptr(k, (T, K), (H*K, 1), (i_t * BT, 0), (BT, BK), (1, 0))
    b_v = tl.load(p_k, boundary_check=(0, 1))
    b_k = (b_v * b_beta[:, None]).to(b_v.dtype)

    b_m = tl.where((o_t[:, None] >= o_t[None, :]) & (m_t[:, None] & m_t[None, :]), exp(b_g[:, None] - b_g[None, :]), 0)
    b_s = tl.dot(b_q_star, tl.trans(b_k)) * b_m
    b_ds = tl.dot(b_dq, tl.trans(b_v))
    b_dv += tl.dot(tl.trans(b_s.to(b_dq.dtype)), b_dq)
    b_dm = b_s * b_ds
    b_dm = tl.where(tl.arange(0, BT)[:, None] >= tl.arange(0, BT)[None, :], b_dm, 0)
    b_dg += tl.sum(b_dm, axis=1)
    b_dg -= tl.sum(b_dm, axis=0)
    b_ds = b_ds * b_m
    b_dk += tl.dot(tl.trans(b_ds.to(b_q_star.dtype)), b_q_star)

    b_h = tl.load(p_h, boundary_check=(0, 1))
    b_dg += tl.sum(tl.dot(b_dq, tl.trans(b_h)) * tl.exp(b_g)[:, None] * b_q_star, axis=1)
    b_dh = tl.load(p_dh, boundary_check=(0, 1))
    b_dk2 = tl.dot(b_v, b_dh.to(b_v.dtype)) * b_gk[:, None]
    b_dg -= tl.sum(b_dk2 * b_k, axis=1)
    b_dg_last += tl.sum(b_dk2 * b_k)
    b_dk += b_dk2
    b_dv += tl.dot(b_k, tl.trans(b_dh).to(b_k.dtype)) * b_gk[:, None]
    b_dh = b_dh * b_h
    b_dg_last += tl.sum(b_dh) * exp(b_g_last)

    p_dk_beta = tl.make_block_ptr(dk_beta, (T, K), (H*K, 1), (i_t * BT, 0), (BT, BK), (1, 0))
    b_dk -= tl.load(p_dk_beta, boundary_check=(0, 1))
    b_dbeta = tl.sum(b_dk * b_v, axis=1)
    b_dk = b_dk * b_beta[:, None] + b_dv
    b_dk = -b_dk

    b_dg = tl.where(o_t < min(i_t * BT + BT, T) - 1, b_dg, b_dg + b_dg_last)
    p_dk = tl.make_block_ptr(dk, (T, K), (H*K, 1), (i_t * BT, 0), (BT, BK), (1, 0))
    tl.store(p_dk, b_dk.to(p_dk.dtype.element_ty), boundary_check=(0, 1))
    p_dg = tl.make_block_ptr(dg, (T,), (H,), (i_t * BT,), (BT,), (0,))
    tl.store(p_dg, -b_dg.to(p_dg.dtype.element_ty), boundary_check=(0,))
    p_dbeta = tl.make_block_ptr(dbeta, (T,), (H,), (i_t * BT,), (BT,), (0,))
    tl.store(p_dbeta, -b_dbeta.to(p_dbeta.dtype.element_ty), boundary_check=(0,))


def chunk_mesa_net_h_kk_bwd_intra_fn(
    k: torch.Tensor,
    beta: torch.Tensor,
    g: torch.Tensor,
    h: torch.Tensor,
    dh: torch.Tensor,
    q_star: torch.Tensor,
    dq: torch.Tensor,
    dk_beta: torch.Tensor,
    cu_seqlens: torch.LongTensor | None = None,
    chunk_size: int = 64,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:

    B, T, H, K = k.shape
    V = K
    BT = min(chunk_size, max(16, triton.next_power_of_2(T)))
    chunk_indices = prepare_chunk_indices(cu_seqlens, BT) if cu_seqlens is not None else None
    NT = triton.cdiv(T, BT) if cu_seqlens is None else len(chunk_indices)

    # CONST_TILING = 64
    BK = max(triton.next_power_of_2(K), 16)
    BV = max(triton.next_power_of_2(V), 16)
    dk = torch.empty_like(k)
    dg = torch.empty_like(g)
    dbeta = torch.empty_like(beta)
    dlamb = torch.empty(B, NT, H, K, dtype=torch.float32, device=k.device)
    grid = (NT, B * H)

    chunk_mesa_net_h_kk_bwd_intra_kernel[grid](
        k=k,
        h=h,
        dh=dh,
        g=g,
        q_star=q_star,
        beta=beta,
        dbeta=dbeta,
        dq=dq,
        dk=dk,
        dk_beta=dk_beta,
        dg=dg,
        dlamb=dlamb,
        cu_seqlens=cu_seqlens,
        chunk_indices=chunk_indices,
        B=B,
        T=T,
        H=H,
        K=K,
        V=V,
        BT=BT,
        BK=BK,
        BV=BV,
    )
    dlamb = dlamb.sum([0, 1])
    return dk, dg, dlamb, dbeta



import triton
from triton import language as tl


@triton.jit
def get_dtype_max(dtype: tl.constexpr):
    """get a value which is greater that all other values of that dtype"""
    # extract the tl.dtype from tl.constexpr so as to use its methods
    dtype_ = dtype.value
    if dtype_.is_floating():
        value: tl.constexpr = float("inf")
        return value
    if dtype_.is_int_signed():
        width: tl.constexpr = dtype_.int_bitwidth
        value: tl.constexpr = 2 ** (width - 1) - 1
        return value
    if dtype_.is_int_unsigned():
        width: tl.constexpr = dtype_.int_bitwidth
        value: tl.constexpr = 2**width - 1
        return value


@triton.jit
def get_dtype_min(dtype):
    """get a value which is less that all other values of that dtype"""
    dtype_ = dtype.value  # tl.dtype
    if dtype_.is_floating():
        value: tl.constexpr = float("-inf")
        return value
    if dtype_.is_int_signed():
        width: tl.constexpr = dtype_.int_bitwidth
        value: tl.constexpr = -1 * 2 ** (width - 1)
        return value
    if dtype_.is_int_unsigned():
        value: tl.constexpr = 0
        return value



import logging

import torch
import triton
import triton.language as tl

from ..utils.pointwise_dynamic import pointwise_dynamic

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


@pointwise_dynamic(is_tensor=[True], promotion_methods=[(0, "ALWAYS_BOOL")])
@triton.jit
def isfinite_func(x):
    if x.dtype.is_fp64():
        return (x.to(tl.int64, bitcast=True) & 0x7FFFFFFFFFFFFFFF) < 0x7FF0000000000000
    elif x.dtype.is_fp32():
        return (x.to(tl.int32, bitcast=True) & 0x7FFFFFFF) < 0x7F800000
    elif x.dtype.is_fp16():
        return (x.to(tl.int16, bitcast=True) & 0x7FFF) < 0x7C00
    elif x.dtype.is_bf16():
        return (x.to(tl.int16, bitcast=True) & 0x7FFF) < 0x7F80


def isfinite(
    A: torch.Tensor,
) -> torch.Tensor:
    logger.debug("GEMS_CAMBRICON ISFINITE")
    if A.is_floating_point():
        legal_dtype = [torch.float32, torch.float16, torch.bfloat16]
        assert (
            A.dtype in legal_dtype
        ), f"isfinite input float dtype should in {str(legal_dtype)}, get {str(A.dtype)}"
        return isfinite_func(A)
    else:
        return torch.full(A.shape, True, dtype=torch.bool, device=A.device)




import torch
import triton
import triton.language as tl

from fla.ops.utils.index import prepare_chunk_indices
from fla.utils import autotune_cache_kwargs


@triton.heuristics({
    'IS_VARLEN': lambda args: args['cu_seqlens'] is not None,
})
@triton.autotune(
    configs=[
        triton.Config({}, num_warps=num_warps)
        for num_warps in [1, 2, 4, 8]
    ],
    key=['B', 'H', 'BT', 'IS_VARLEN'],
    **autotune_cache_kwargs,
)
@triton.jit(do_not_specialize=['T'])
def chunk_comba_cumsum_scalar_fwd_kernel(
    g,
    g0,
    g1,
    cu_seqlens,
    chunk_indices,
    T,
    B: tl.constexpr,
    H: tl.constexpr,
    BT: tl.constexpr,
    IS_VARLEN: tl.constexpr,
    HEAD_FIRST: tl.constexpr,
):
    i_t, i_bh = tl.program_id(0), tl.program_id(1)
    i_b, i_h = i_bh // H, i_bh % H
    if IS_VARLEN:
        i_n, i_t = tl.load(chunk_indices + i_t * 2).to(tl.int32), tl.load(chunk_indices + i_t * 2 + 1).to(tl.int32)
        bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(cu_seqlens + i_n + 1).to(tl.int32)
        T = eos - bos
    else:
        bos, eos = i_b * T, i_b * T + T

    if HEAD_FIRST:
        p_g = tl.make_block_ptr(g + bos*H + i_h*T, (T,), (1,), (i_t * BT,), (BT,), (0,))
        p_g0 = tl.make_block_ptr(g0 + bos*H + i_h*T, (T,), (1,), (i_t * BT,), (BT,), (0,))
        p_g1 = tl.make_block_ptr(g1 + bos*H + i_h*T, (T,), (1,), (i_t * BT,), (BT,), (0,))
    else:
        p_g = tl.make_block_ptr(g + bos*H + i_h, (T,), (H,), (i_t * BT,), (BT,), (0,))
        p_g0 = tl.make_block_ptr(g0 + bos*H + i_h, (T,), (H,), (i_t * BT,), (BT,), (0,))
        p_g1 = tl.make_block_ptr(g1 + bos*H + i_h, (T,), (H,), (i_t * BT,), (BT,), (0,))
    # [BT]
    b_g = tl.load(p_g, boundary_check=(0,)).to(tl.float32)
    b_g1 = tl.cumsum(b_g, axis=0)
    b_g0 = b_g1 - b_g
    tl.store(p_g0, b_g0.to(p_g0.dtype.element_ty), boundary_check=(0,))
    tl.store(p_g1, b_g1.to(p_g1.dtype.element_ty), boundary_check=(0,))


def chunk_comba_cumsum_scalar_fwd(
    g: torch.Tensor,
    chunk_size: int,
    cu_seqlens: torch.Tensor | None = None,
    head_first: bool = False,
    output_dtype: torch.dtype | None = torch.float,
) -> torch.Tensor:
    if head_first:
        B, H, T = g.shape
    else:
        B, T, H = g.shape
    assert chunk_size == 2**(chunk_size.bit_length()-1), "chunk_size must be a power of 2"
    BT = chunk_size
    chunk_indices = prepare_chunk_indices(cu_seqlens, BT) if cu_seqlens is not None else None
    NT = triton.cdiv(T, BT) if cu_seqlens is None else len(chunk_indices)
    g0, g1 = torch.empty_like(g, dtype=output_dtype or g.dtype), torch.empty_like(g, dtype=output_dtype or g.dtype)
    grid = (NT, B * H)
    chunk_comba_cumsum_scalar_fwd_kernel[grid](
        g,
        g0,
        g1,
        cu_seqlens,
        chunk_indices,
        T=T,
        B=B,
        H=H,
        BT=BT,
        HEAD_FIRST=head_first,
    )
    return g0, g1


@triton.heuristics({
    'IS_VARLEN': lambda args: args['cu_seqlens'] is not None,
})
@triton.autotune(
    configs=[
        triton.Config({}, num_warps=num_warps)
        for num_warps in [1, 2, 4, 8]
    ],
    key=['B', 'H', 'BT', 'IS_VARLEN'],
    **autotune_cache_kwargs,
)
@triton.jit(do_not_specialize=['T'])
def chunk_comba_cumsum_scalar_bwd_kernel(
    dg0,
    dgr,
    cu_seqlens,
    chunk_indices,
    T,
    B: tl.constexpr,
    H: tl.constexpr,
    BT: tl.constexpr,
    IS_VARLEN: tl.constexpr,
    HEAD_FIRST: tl.constexpr,
):
    i_t, i_bh = tl.program_id(0), tl.program_id(1)
    i_b, i_h = i_bh // H, i_bh % H
    if IS_VARLEN:
        i_n, i_t = tl.load(chunk_indices + i_t * 2).to(tl.int32), tl.load(chunk_indices + i_t * 2 + 1).to(tl.int32)
        bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(cu_seqlens + i_n + 1).to(tl.int32)
        T = eos - bos
    else:
        bos, eos = i_b * T, i_b * T + T

    if HEAD_FIRST:
        p_dg0 = tl.make_block_ptr(dg0 + bos*H + i_h*T, (T,), (1,), (i_t * BT,), (BT,), (0,))
        p_dgr = tl.make_block_ptr(dgr + bos*H + i_h*T, (T,), (1,), (i_t * BT,), (BT,), (0,))
    else:
        p_dg0 = tl.make_block_ptr(dg0 + bos*H + i_h, (T,), (H,), (i_t * BT,), (BT,), (0,))
        p_dgr = tl.make_block_ptr(dgr + bos*H + i_h, (T,), (H,), (i_t * BT,), (BT,), (0,))
    # [BT]
    """
    b_dg:   1,2,3,4
    b_dg0:  0,1,2,3
    b_temp: 0,1,3,6
    b_dz:   6
    b_dgr:  6,5,3,0
    """
    b_dg0 = tl.load(p_dg0, boundary_check=(0,)).to(tl.float32)
    b_temp = tl.cumsum(b_dg0, axis=0)
    b_dz = tl.sum(b_dg0, axis=0)
    b_dgr = -b_temp + b_dz[None]
    tl.store(p_dgr, b_dgr.to(p_dgr.dtype.element_ty), boundary_check=(0,))


def chunk_comba_cumsum_scalar_bwd(
    dg0: torch.Tensor,
    chunk_size: int,
    cu_seqlens: torch.Tensor | None = None,
    head_first: bool = False,
    output_dtype: torch.dtype | None = torch.float,
) -> torch.Tensor:
    if head_first:
        B, H, T = dg0.shape
    else:
        B, T, H = dg0.shape
    assert chunk_size == 2**(chunk_size.bit_length()-1), "chunk_size must be a power of 2"
    BT = chunk_size
    chunk_indices = prepare_chunk_indices(cu_seqlens, BT) if cu_seqlens is not None else None
    NT = triton.cdiv(T, BT) if cu_seqlens is None else len(chunk_indices)
    dg = torch.empty_like(dg0, dtype=output_dtype or dg0.dtype)
    grid = (NT, B * H)
    chunk_comba_cumsum_scalar_bwd_kernel[grid](
        dg0,
        dg,
        cu_seqlens,
        chunk_indices,
        T=T,
        B=B,
        H=H,
        BT=BT,
        HEAD_FIRST=head_first,
    )
    return dg



import logging
import math
from collections import namedtuple

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.runtime import torch_device_fn
from flag_gems.utils import libentry, libtuner
from flag_gems.utils.limits import get_dtype_min

from ..utils import TOTAL_CORE_NUM, cfggen_reduce_op, prune_reduce_config

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


@libentry()
@triton.jit
def max_kernel_float_once(
    inp,
    out,
    M: tl.constexpr,
):
    offset = tl.arange(0, M)
    inp_val = tl.load(inp + offset)
    max_val = tl.max(inp_val, 0)
    tl.store(out, max_val)


@libentry()
@libtuner(
    configs=cfggen_reduce_op(),
    key=["M"],
    strategy=["log"],
    prune_configs_by={"early_config_prune": prune_reduce_config},
)
@triton.heuristics(
    values={
        "ONE_TILE_PER_CTA": lambda args: args["M"]
        <= args["BLOCK_SIZE"] * TOTAL_CORE_NUM
    }
)
@triton.jit
def max_kernel_float(
    inp, out, M, BLOCK_SIZE: tl.constexpr, ONE_TILE_PER_CTA: tl.constexpr
):
    pid = tl.program_id(0)
    block_start = pid * BLOCK_SIZE
    res = -float("inf")

    if ONE_TILE_PER_CTA:
        offset = block_start + tl.arange(0, BLOCK_SIZE)
        mask = offset < M
        inp_val = tl.load(inp + offset, mask=mask, other=-float("inf"))
        (res,) = tl.max(inp_val, 0, return_indices=True)
        tl.atomic_max(out, res)
    else:
        num_jobs = tl.num_programs(axis=0)
        step = num_jobs * BLOCK_SIZE
        _tmp = tl.full([BLOCK_SIZE], value=-float("inf"), dtype=inp.dtype.element_ty)
        for off in range(block_start, M, step):
            offset = off + tl.arange(0, BLOCK_SIZE)
            mask = offset < M
            inp_val = tl.load(inp + offset, mask=mask, other=-float("inf"))
            _tmp = tl.where((inp_val > _tmp), inp_val, _tmp)
        (res,) = tl.max(_tmp, 0, return_indices=True)
        tl.atomic_max(out, res)


@libentry()
@libtuner(
    configs=cfggen_reduce_op(),
    key=["M"],
    strategy=["log"],
    prune_configs_by={"early_config_prune": prune_reduce_config},
)
@triton.heuristics(
    values={
        "ONE_TILE_PER_CTA": lambda args: args["M"]
        <= args["BLOCK_SIZE"] * TOTAL_CORE_NUM
    }
)
@triton.jit
def max_kernel_int(
    inp, out, M, FILL_VALUE, BLOCK_SIZE: tl.constexpr, ONE_TILE_PER_CTA: tl.constexpr
):
    pid = tl.program_id(0)
    block_start = pid * BLOCK_SIZE
    res = FILL_VALUE
    if ONE_TILE_PER_CTA:
        offset = block_start + tl.arange(0, BLOCK_SIZE)
        mask = offset < M
        inp_val = tl.load(inp + offset, mask=mask, other=FILL_VALUE)
        res = tl.max(inp_val)
    else:
        num_jobs = tl.num_programs(axis=0)
        step = num_jobs * BLOCK_SIZE
        block_start = pid * BLOCK_SIZE
        _tmp = tl.full([BLOCK_SIZE], value=-(2**63), dtype=tl.int64)
        for off in range(block_start, M, step):
            offset = off + tl.arange(0, BLOCK_SIZE)
            mask = offset < M
            inp_val = tl.load(inp + offset, mask=mask, other=FILL_VALUE)
            _tmp = tl.where((inp_val > _tmp), inp_val, _tmp)
        res = tl.max(_tmp)
    tl.atomic_max(out, res)


@libentry()
@libtuner(
    configs=cfggen_reduce_op(),
    key=["M"],
    strategy=["log"],
    prune_configs_by={"early_config_prune": prune_reduce_config},
)
@triton.heuristics(
    values={
        "ONE_TILE_PER_CTA": lambda args: args["M"]
        <= args["BLOCK_SIZE"] * TOTAL_CORE_NUM
    }
)
@triton.jit
def max_kernel_int64_1(
    inp, mid, M, BLOCK_SIZE: tl.constexpr, ONE_TILE_PER_CTA: tl.constexpr
):
    pid = tl.program_id(0)
    block_start = pid * BLOCK_SIZE
    FILL_VALUE = -(2**63)
    res = FILL_VALUE
    if ONE_TILE_PER_CTA:
        offset = block_start + tl.arange(0, BLOCK_SIZE)
        mask = offset < M
        inp_val = tl.load(inp + offset, mask=mask, other=FILL_VALUE)
        res = tl.max(inp_val)
    else:
        num_jobs = tl.num_programs(axis=0)
        step = num_jobs * BLOCK_SIZE
        block_start = block_start.to(tl.int64)
        _tmp = tl.full([BLOCK_SIZE], value=FILL_VALUE, dtype=tl.int64)
        for off in range(block_start, M, step):
            offset = off + tl.arange(0, BLOCK_SIZE)
            mask = offset < M
            inp_val = tl.load(inp + offset, mask=mask, other=FILL_VALUE)
            _tmp = tl.where((inp_val > _tmp), inp_val, _tmp)
        res = tl.max(_tmp)
    tl.store(mid + pid, res)


@libentry()
@triton.jit
def max_kernel_int64_2(mid, out, BLOCK_NUM: tl.constexpr):
    offset = tl.arange(0, BLOCK_NUM)
    mid_val = tl.load(mid + offset)
    out_val = tl.max(mid_val)
    tl.store(out, out_val)


def heur_block_n(args):
    return triton.next_power_of_2(args["N"])


@libentry()
@libtuner(
    configs=runtime.get_tuned_config("max"),
    key=[
        "M",
        "N",
    ],
    strategy=["log", "log"],
)
@triton.jit
def max_kernel(
    inp,
    out_value,
    out_index,
    M,
    N,
    K,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    # set offset
    pid_m = tl.program_id(0)
    pid_k = tl.program_id(1)
    m_offset = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    result_value = tl.full([BLOCK_M], value=-float("inf"), dtype=tl.float32)
    result_index = tl.zeros([BLOCK_M], dtype=tl.int64)
    min_value = get_dtype_min(inp.type.element_ty)
    for i in range(0, N, BLOCK_N):
        n_offset = i + tl.arange(0, BLOCK_N)
        offset = m_offset[:, None] * N * K + n_offset[None, :] * K + pid_k
        # set mask
        mask = m_offset[:, None] < M and n_offset[None, :] < N
        inp_ptrs = inp + offset
        inp_vals = tl.load(inp_ptrs, mask=mask, other=min_value)
        max_value, max_index = tl.max(inp_vals, axis=1, return_indices=True)
        update_mask = max_value > result_value
        result_value = tl.where(update_mask, max_value, result_value)
        result_index = tl.where(update_mask, i + max_index, result_index)
    mask1 = m_offset < M
    offset_index = m_offset * K + pid_k
    out_value_ptrs = out_value + offset_index
    out_index_ptrs = out_index + offset_index

    tl.store(out_value_ptrs, result_value, mask=mask1)
    tl.store(out_index_ptrs, result_index, mask=mask1)


def max(inp):
    logger.debug("GEMS_CAMBRICON MAX")
    inp = inp.contiguous()
    M = inp.numel()
    dtype = inp.dtype
    device = inp.device
    mid_size = TOTAL_CORE_NUM
    grid = lambda meta: (min(triton.cdiv(M, meta["BLOCK_SIZE"]), mid_size),)

    with torch_device_fn.device(inp.device):
        if torch.is_floating_point(inp):
            if M <= 65536:
                out = torch.empty([], dtype=dtype, device=device)
                max_kernel_float_once[(1, 1, 1)](inp, out, M)
            else:
                out = torch.full([], float("-inf"), dtype=torch.float32, device=device)
                max_kernel_float[grid](inp, out, M)
        elif dtype == torch.int64:
            mid = torch.empty([mid_size], dtype=dtype, device=device)
            out = torch.empty([], dtype=dtype, device=device)
            # Because atomic op don't support i64, use two kernels.
            max_kernel_int64_1[(mid_size, 1, 1)](inp, mid, M, enable_soft_i64=True)
            max_kernel_int64_2[(1, 1, 1)](
                mid, out, BLOCK_NUM=mid_size, enable_soft_i64=True
            )
        else:
            fill_value = torch.iinfo(dtype).min
            out = torch.full([], -(2**31), dtype=torch.int32, device=device)
            max_kernel_int[grid](inp, out, M, fill_value)
    return out.to(dtype)


def max_dim(inp, dim=None, keepdim=False):
    logger.debug("GEMS_CAMBRICON MAX DIM")
    assert dim >= -inp.ndim and dim < inp.ndim, "Invalid dim"
    shape = inp.shape
    dim = dim % inp.ndim
    N = shape[dim]
    M = math.prod(shape[:dim])
    K = inp.numel() // M // N

    inp = inp.contiguous()

    shape_list = list(shape)
    shape_list[dim] = 1
    out_value = torch.empty(shape_list, dtype=inp.dtype, device=inp.device)
    out_index = torch.empty(shape_list, dtype=torch.int64, device=inp.device)

    if not keepdim:
        out_value = torch.squeeze(out_value, dim)
        out_index = torch.squeeze(out_index, dim)

    grid = lambda meta: (
        triton.cdiv(M, meta["BLOCK_M"]),
        K,
    )
    with torch_device_fn.device(inp.device):
        max_kernel[grid](inp, out_value, out_index, M, N, K)
    Max_out = namedtuple("max", ["values", "indices"])
    out = Max_out(values=out_value, indices=out_index)
    return out



import importlib
import logging
import math
import os
from typing import Callable, List, Mapping, Tuple, Union

import torch

from flag_gems.utils.code_cache import code_cache_dir
from flag_gems.utils.code_utils import IndentedBuffer, write_atomic

from .vstack import vstack

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


class CatKernelGenerator(IndentedBuffer):
    overloads: Mapping[str, Callable] = {}

    def __init__(self):
        self.pid = os.getpid()
        self.cache = self.overloads
        super().__init__()

    def __init(
        self,
        tensors: List[torch.Tensor],
        dim: int,
        high_num: int,
        low_cat_accum: List[int],
    ):
        self.dim = dim
        self.high_num = high_num
        self.low_cat_accum = low_cat_accum
        self.tensor_num = len(tensors)
        even = all([t.numel() == tensors[0].numel() for t in tensors])

        if even and low_cat_accum[-1] // self.tensor_num <= 128:
            # Special case for tensors with small and even low size,
            # which means weak contiguity when storing the out tensor.
            # Divide each tensor into tiles of `BLOCK_LOW` size,
            # and each cta process tiles one by one.
            self.kernel_name = "_cat_kernel_small"
            self.wrapper_name = "_cat_wrapper_small"
            self.MODE = 0
        else:
            # General cases.
            # Divide tasks by high_num, each cta process parts of high of all tensors.
            self.kernel_name = "_cat_kernel_parthigh"
            self.wrapper_name = "_cat_wrapper_parthigh"
            self.MODE = 1

    def __call__(
        self,
        tensors: List[torch.Tensor],
        dim: int,
        high_num: int,
        low_cat_accum: List[int],
    ):
        self.__init(tensors, dim, high_num, low_cat_accum)
        key = f"{len(tensors)}_{high_num}_{low_cat_accum[-1]}"
        if key not in self.cache:
            self.codegen()

            filename = f"{self.kernel_name}_{key}.py"
            filepath = code_cache_dir() / filename
            write_atomic(filepath, self.getvalue())

            spec = importlib.util.spec_from_file_location(
                f"_gen_module_{key}", filepath
            )
            m = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(m)
            overload = getattr(m, self.wrapper_name)
            self.cache[key] = overload
        overload = self.cache[key]
        return overload(tensors, dim, high_num, low_cat_accum)

    def gen_imports(self):
        self.writeline("import math")
        self.writeline("import copy")
        self.newline()
        self.writeline("import torch")
        self.writeline("import triton")
        self.writeline("import triton.language as tl")
        self.newline()
        self.writeline("from flag_gems.runtime import torch_device_fn")
        self.writeline("from flag_gems.runtime.backend import vendor_module")
        self.writeline("from flag_gems.utils import libentry, libtuner")
        self.newline()
        self.writeline("TOTAL_CORE_NUM = vendor_module.TOTAL_CORE_NUM")
        self.newline()
        self.newline()

    def gen_wrapper(self):
        self.writeline(
            f"def {self.wrapper_name}(tensors, dim, high_num, low_cat_accum):"
        )
        with self.indent():
            self.writeline("device = tensors[0].device")
            self.writeline("dtype = tensors[0].dtype")
            self.writeline("tensor_num = len(tensors)")
            self.writeline("cat_dim_size = sum([t.shape[dim] for t in tensors])")
            self.writeline("out_shape = list(tensors[0].shape)")
            self.writeline("out_shape[dim] = cat_dim_size")
            self.writeline("out_cat_num = low_cat_accum[-1]")
            self.writeline("out = torch.empty(out_shape, device=device, dtype=dtype)")
            for i in range(self.tensor_num):
                self.writeline(f"in{i}_stride_high = tensors[{i}].stride(dim - 1)")
                self.writeline(f"in{i}_stride_low = tensors[{i}].stride(-1)")
            self.writeline("out_stride_high = out.stride(dim - 1)")
            self.writeline("out_stride_low = out.stride(-1)")
            self.writeline(
                "grid = lambda meta: (TOTAL_CORE_NUM // meta['num_warps'], )"
            )
            self.writeline("with torch_device_fn.device(device):")
            with self.indent():
                self.writeline(
                    f"{self.kernel_name}[grid]({self.gen_kernel_args(is_declare=False)})"
                )
            self.writeline("return out")
        self.newline()
        self.newline()

    def gen_decorators(self):
        self.writeline("@libentry()")
        self.writeline("@libtuner(")
        with self.indent():
            self.writeline("configs=[")
            with self.indent():
                if self.MODE == 0:
                    self.writeline(
                        """
        triton.Config({'BLOCK_LOW': 2 ** i}, num_stages=1, num_warps=1) for i in range(7, 12)
                        """
                    )
                elif self.MODE == 1:
                    self.writeline(
                        """
        triton.Config({'BLOCK_HIGH': i, 'BLOCK_LOW': 2 ** j}, num_stages=1, num_warps=1)
        for i in [6, 11, 22]
        for j in range(8, 12)
                        """
                    )
                self.writeline("],")
            self.writeline("key=['high_num', 'out_cat_num'],")
            self.writeline("strategy=['log', 'log'],")
            self.writeline("restore_value=['out'],")
        self.writeline(")")
        self.writeline("@triton.jit")

    def gen_kernel(self):
        self.writeline(f"def {self.kernel_name}({self.gen_kernel_args()}):")
        with self.indent():
            self.writeline("pid = tl.program_id(0)")
            self.writeline("programs_num = tl.num_programs(0)")
            if self.MODE == 0:
                self.writeline(
                    "tiles_per_tensor = tl.cdiv(high_num * tl.cdiv(out_cat_num, tensor_num), BLOCK_LOW)"
                )
                self.writeline("num_tiles = tiles_per_tensor * tensor_num")
                self.writeline("tiles_per_cta = tl.cdiv(num_tiles, programs_num)")
                self.writeline("for i in range(tiles_per_cta):")
                with self.indent():
                    self.writeline("tile_id = pid + i * programs_num")
                    self.writeline("tensor_id = tile_id // tiles_per_tensor")
                    self.writeline("tile_id = tile_id % tiles_per_tensor")
                    for j in range(self.tensor_num):
                        self.writeline(f"if tensor_id == {j}:")
                        with self.indent():
                            self.writeline(
                                f"low_cat = low_cat_accum{j + 1} - low_cat_accum{j}"
                            )
                            self.writeline("offsets = tl.arange(0, BLOCK_LOW)")
                            self.writeline("in_offsets = tile_id * BLOCK_LOW + offsets")
                            self.writeline("mask = in_offsets < high_num * low_cat")
                            self.writeline(
                                f"data = tl.load(in{j} + in_offsets, mask=mask)"
                            )
                            high_part = "(in_offsets // low_cat) * out_cat_num"
                            low_part = f"low_cat_accum{j} + (in_offsets % low_cat)"
                            self.writeline(f"out_offsets = {high_part} + {low_part}")
                            self.writeline(
                                "tl.store(out + out_offsets, data, mask=mask)"
                            )
            elif self.MODE == 1:
                self.writeline("num_tiles = tl.cdiv(high_num, BLOCK_HIGH)")
                self.writeline("tiles_per_cta = tl.cdiv(num_tiles, programs_num)")
                self.writeline("for i in range(tiles_per_cta):")
                with self.indent():
                    self.writeline("tile_id = pid + i * programs_num")
                    self.writeline("high_offset = tile_id * BLOCK_HIGH")
                    for j in range(self.tensor_num):
                        self.writeline(
                            f"low_cat = low_cat_accum{j + 1}-low_cat_accum{j}"
                        )
                        self.writeline(
                            "for low_offset in range(0, low_cat, BLOCK_LOW):"
                        )
                        with self.indent():
                            self.writeline(
                                "high_offsets = high_offset + tl.arange(0, BLOCK_HIGH)"
                            )
                            self.writeline(
                                "low_offsets = low_offset + tl.arange(0, BLOCK_LOW)"
                            )
                            high_part = f"high_offsets[:, None] * in{j}_stride_high"
                            low_part = f"low_offsets[None, :] * in{j}_stride_low"
                            self.writeline(f"in_offsets = {high_part} + {low_part}")
                            self.writeline(
                                "in_mask = (high_offsets < high_num)[:,None] & (low_offsets < low_cat)[None,:]"
                            )
                            self.writeline(
                                f"data = tl.load(in{j}+in_offsets, mask=in_mask)"
                            )
                            high_part = "high_offsets[:, None] * out_stride_high"
                            low_part = f"(low_cat_accum{j} + low_offsets[None, :]) * out_stride_low"
                            self.writeline(f"out_offsets = {high_part} + {low_part}")
                            self.writeline(
                                "tl.store(out+out_offsets, data, mask=in_mask)"
                            )

    def gen_kernel_args(self, is_declare=True):
        in_args = ", ".join(
            f"in{i}" if is_declare else f"tensors[{i}]" for i in range(self.tensor_num)
        )
        low_cat_accum_args = ", ".join(
            f"low_cat_accum{i}" if is_declare else f"low_cat_accum[{i}]"
            for i in range(self.tensor_num + 1)
        )
        stride_args = (
            ", ".join(
                f"in{i}_stride_high, in{i}_stride_low" for i in range(self.tensor_num)
            )
            + ", out_stride_high, out_stride_low"
        )

        kernel_args = f"{in_args}, out, {stride_args}, tensor_num, high_num, {low_cat_accum_args}, out_cat_num, "
        ex_args = "BLOCK_LOW: tl.constexpr, num_warps: tl.constexpr"
        if self.MODE == 1:
            ex_args += ", BLOCK_HIGH: tl.constexpr"

        return kernel_args if not is_declare else kernel_args + ex_args

    def codegen(self):
        self.gen_imports()
        self.gen_wrapper()
        self.gen_decorators()
        self.gen_kernel()


def cat(
    tensors: Union[Tuple[torch.Tensor, ...], List[torch.Tensor]], dim: int = 0
) -> torch.Tensor:
    logger.debug("GEMS_CAMBRICON CAT")

    # Check empty inputs.
    if len(tensors) == 0:
        raise RuntimeError(
            "Expected a non-empty list or tuple/list of non-empty torch.Tensor"
        )

    # Check dimensions.
    ndim = tensors[0].ndim
    assert dim >= -ndim and dim < ndim, f"Invalid concat dimension: {dim}"
    dim %= ndim

    # Check shapes and zero element tensors.
    device = tensors[0].device
    dtype = tensors[0].dtype
    shape = tensors[0].shape
    valid_tensors = []

    for _, tensor in enumerate(tensors):
        assert (
            tensor.ndim == ndim
        ), f"Requires same ndim of inputs, but got {ndim} and {tensor.ndim}"
        assert (
            tensor.device == device
        ), f"Requires same device of inputs, but got {device} and {tensor.device}"
        assert (
            tensor.dtype == dtype
        ), f"Requires same dtype of inputs, but got {dtype} and {tensor.dtype}"
        if tensor.numel() != 0:
            valid_tensors.append(tensor.contiguous())
        for d_idx, (size, base_size) in enumerate(zip(tensor.shape, shape)):
            assert (
                dim == d_idx or size == base_size
            ), f"Requires same dim sizes of dim {d_idx}, but got {size} and {base_size}"

    tensor_num = len(valid_tensors)

    # Deal with special cases.
    if tensor_num == 1:
        return valid_tensors[0]

    cat_dim_sizes = [_.shape[dim] for _ in tensors]
    out_shape = list(tensors[0].shape)
    out_shape[dim] = sum(cat_dim_sizes)

    if tensor_num == 0:
        return torch.empty(out_shape, dtype=dtype, device=device)

    # Preprocess kernel parameters.
    high_num = int(math.prod(out_shape[:dim]))
    low_num = int(math.prod(out_shape[dim + 1 :]))
    out_cat_num = 0
    low_cat_accum = [0]

    for size in cat_dim_sizes:
        out_cat_num += size * low_num
        low_cat_accum.append(out_cat_num)

    # Launch kernel.
    if high_num == 1:
        # Vstack and Concat results in the same storage arrangement when high_num == 1.
        valid_tensors = [t.view(t.shape[dim], -1) for t in valid_tensors]
        return vstack(valid_tensors).view(out_shape)
    else:
        # Dealing with concat situations that having arbitary nums of inputs via template code genertaor.
        return CatKernelGenerator()(valid_tensors, dim, high_num, low_cat_accum)
